#include "bch.h"

#include <stdlib.h>

#include "matrice.h"
#include "peterson_solve.h"
#include "construct_poly.h"

static int get_order(int val, int mod)
{
	int order = 0;
	int base = val;
	
	do
	{
		val *= 2;
		val %= mod;
		++order;
	} while(base != val);
	
	return order;
}

static int solve(GenPolyF2 * res, int power, int order, const GenPolyF2 * based_poly)
{
	MatriceF2 reference;
	MatriceF2 trv;
	GenPolyF2 racine;
	GenPolyF2 add;
	int i;
	
	if(! GenPolyF2_create_single(res, PT_NORMAL, order))
		return 0;
	if(! GenPolyF2_create_single(&racine, PT_NORMAL, power))
	{
		GenPolyF2_free(res);
		return 0;
	}
	if(! GenPolyF2_create_single(&add, PT_NORMAL, 0))
	{
		GenPolyF2_free(res);
		GenPolyF2_free(&racine);
		return 0;
	}
	if(! MatriceF2_create(&reference, GenPolyF2_real_degree(based_poly), order))
	{
		GenPolyF2_free(res);
		GenPolyF2_free(&racine);
		GenPolyF2_free(&add);
		return 0;
	}
	
	for(i = 0; i < order; ++i)
	{
		if(! MatriceF2_poly_to_colonne(&reference, &add, i))
			goto error_zone;
		GenPolyF2_free(&add);
		if(! GenPolyF2_power(&add, &racine, i + 1, based_poly))
		{
			GenPolyF2_free(res);
			GenPolyF2_free(&racine);
			return 0;
		}
	}
	
	for(i = 0; i < order; ++i)
	{
		if(! MatriceF2_copy(&trv, &reference))
			goto error_zone;
		if(! MatriceF2_poly_to_colonne(&trv, &add, i)
		|| ! MatriceF2_apply_gauss_colonne(&trv, NULL))
		{
			MatriceF2_free(&trv);
			goto error_zone;
		}
		if(! MatriceF2_is_null_colonne(&trv))
		{
			if(! GenPolyF2_set_val(res, 1, i))
			{
				MatriceF2_free(&trv);
				goto error_zone;
			}
		}
		MatriceF2_free(&trv);
	}
	
	MatriceF2_free(&reference);
	GenPolyF2_free(&racine);
	GenPolyF2_free(&add);
	
	return 1;
	
	error_zone:;
	
	MatriceF2_free(&reference);
	GenPolyF2_free(res);
	GenPolyF2_free(&racine);
	GenPolyF2_free(&add);
	
	return 0;
}

static int solve_c(GenPolyF2 * res, int power, int len, const GenPolyF2 * based_poly)
{
	CPoly dev;
	CPoly pracine;
	GenPolyF2 racine[2];
	GenPolyF2 alpha;
	int order;
	int i;
	
	if(! GenPolyF2_create_single(&alpha, PT_NORMAL, power))
		return 0;
	if(! GenPolyF2_create_single(racine + 1, PT_NORMAL, 0))
	{
		GenPolyF2_free(&alpha);
		return 0;
	}
	if(! CPoly_create_by_tab(&dev, racine + 1, 1, based_poly))
	{
		GenPolyF2_free(&alpha);
		GenPolyF2_free(racine + 1);
		return 0;
	}
	order = get_order(power, len);
	for(i = 0; i < order; ++i)
	{
		if(! GenPolyF2_power(racine, &alpha, i + 1, based_poly))
			goto error_zone;
		if(! CPoly_create_by_tab(&pracine, racine, 2, based_poly))
		{
			GenPolyF2_free(racine);
			goto error_zone;
		}
		GenPolyF2_free(racine);
		if(! CPoly_self_mul(&dev, &pracine))
		{
			CPoly_free(&pracine);
			goto error_zone;
		}
		CPoly_free(&pracine);
	}
	if(! CPoly_to_F2(&dev, res))
	{
		/*CPoly_print(&dev);*/
		goto error_zone;
	}
	CPoly_free(&dev);
	GenPolyF2_free(&alpha);
	GenPolyF2_free(racine + 1);
	
	return 1;
	
	error_zone:;
	
	CPoly_free(&dev);
	GenPolyF2_free(&alpha);
	GenPolyF2_free(racine + 1);
	return 0;
}

static int add_class(GenPolyF2 * g, int power, int len, const GenPolyF2 * based_poly)
{
	GenPolyF2 gcd;
	GenPolyF2 res;
	GenPolyF2 tmp;
	
	if(power == 1)
		return GenPolyF2_self_mul(g, based_poly, NULL);
	if(! solve_c(&tmp, power, len, based_poly))
	{
		if(! solve(&tmp, power, get_order(power, len), based_poly))
			return 0;
	}
	if(! GenPolyF2_gcd(&gcd, NULL, NULL, &tmp, g))
	{
		GenPolyF2_free(&tmp);
		return 0;
	}
	if(! GenPolyF2_euclid_div(&res, NULL, &tmp, &gcd))
	{
		GenPolyF2_free(&gcd);
		GenPolyF2_free(&tmp);
	}
	GenPolyF2_free(&gcd);
	GenPolyF2_free(&tmp);
	if(! GenPolyF2_self_mul(g, &res, NULL))
	{
		GenPolyF2_free(&res);
		return 0;
	}
	GenPolyF2_free(&res);
	return 1;
}

int BCH_create(BCH * code, const GenPolyF2 * constructor, int nb_correct)
{
	int len;
	int base_degree;
	GenPolyF2 g;
	GenPolyF2 ref;
	int i;
	
	base_degree = GenPolyF2_real_degree(constructor);
	/*if(nb_correct > max_correct(base_degree))
	{
		printf("Polynome de degré %d trop petit pour corriger %d erreurs, max : (%d)\n", base_degree, nb_correct, max_correct(base_degree));
		return 0;
	}*/
	len = (1 << base_degree) - 1;
	if(! GenPolyF2_create_by_vals(&g, PT_NORMAL, 1, -1))
		return 0;
	for(i = 1; i < 2 * nb_correct; i += 2)
	{
		if(! add_class(&g, i, len, constructor))
		{
			GenPolyF2_free(&g);
			return 0;
		}
	}
	if(! GenPolyF2_create_single(&ref, PT_NORMAL, len))
	{
		GenPolyF2_free(&g);
		return 0;
	}
	GenPolyF2_set_val(&ref, 1, 0);
	if(! GenPolyF2_euclid_div(&(code->h), NULL, &ref, &g))
	{
		GenPolyF2_free(&g);
		GenPolyF2_free(&ref);
		return 0;
	}
	GenPolyF2_free(&ref);
	if(! GenPolyF2_copy(&(code->based_poly), constructor))
	{
		GenPolyF2_free(&g);
		GenPolyF2_free(&(code->h));
		return 0;
	}
	code->g = g;
	code->len = len;
	code->nb_correct = nb_correct;
	code->msg_len = GenPolyF2_real_degree(&(code->h));
	
	return 1;
}

void BCH_free(BCH * code)
{
	GenPolyF2_free(&(code->g));
	GenPolyF2_free(&(code->h));
	GenPolyF2_free(&(code->based_poly));
}

void BCH_print(const BCH * code)
{
	printf("BCH longueur %d pour %d corrections\n", code->len, code->nb_correct);
	printf("messages de taille %d\n", code->msg_len);
	printf("Dans le corps fini F%d généré par F2/", 1 << GenPolyF2_real_degree(&(code->based_poly)));
	GenPolyF2_print_bin(&(code->based_poly));
	printf("g = ");
	GenPolyF2_print_bin(&(code->g));
	printf("h = ");
	GenPolyF2_print_bin(&(code->h));
}

int BCH_code(const BCH * code, GenPolyF2 * res, const GenPolyF2 * message)
{
	GenPolyF2 decal;
	GenPolyF2 reminder;
	
	if(! GenPolyF2_decal(&decal, message, code->len - code->msg_len))
		return 0;
	if(! GenPolyF2_euclid_div(NULL, &reminder, &decal, &(code->g)))
	{
		GenPolyF2_free(&decal);
		return 0;
	}
	if(! GenPolyF2_add(res, &reminder, &decal))
	{
		GenPolyF2_free(&decal);
		GenPolyF2_free(&reminder);
		return 0;
	}
	GenPolyF2_free(&decal);
	GenPolyF2_free(&reminder);
	return 1;
}

static int get_syndroms(const BCH * code, const GenPolyF2 * message, GenPolyF2 ** syndroms, int nb_syndroms)
{
	GenPolyF2 * tab = NULL;
	GenPolyF2 tmp;
	int i;
	
	if((tab = (GenPolyF2 *)calloc(nb_syndroms, sizeof(GenPolyF2))) == NULL)
		return 0;
	for(i = 0; i < nb_syndroms; ++i)
	{
		if(! GenPolyF2_create_single(&tmp, PT_NORMAL, i + 1))
			goto error_zone;
		if(! GenPolyF2_compose(tab + i, message, &tmp, &(code->based_poly)))
		{
			GenPolyF2_free(&tmp);
			goto error_zone;
		}
		GenPolyF2_free(&tmp);
	}
	
	*syndroms = tab;
	
	return 1;
	
	error_zone:;
	
	for(--i; i >= 0; --i)
		GenPolyF2_free(tab + i);
	free(tab);
	
	return 0;
}

static int get_racines(const BCH * code, GenPolyF2 * syndroms, CPoly * racines)
{
	return CPoly_locator(racines, &(code->based_poly), syndroms, code->nb_correct);
}

/* Fonctionne pour nb_correct == 2 mais non pour nb_correct == 3 */
#if 0
static int get_racines(const BCH * code, GenPolyF2 * syndroms, GenPolyF2 ** racines)
{
	GenPolyF2 s1_inv;
	GenPolyF2 * tab = NULL;
	GenPolyF2 sum;
	GenPolyF2 tmp;
	int i;
	int j;
	
	if(! GenPolyF2_invert(&s1_inv, syndroms, &(code->based_poly)))
		return 0;
	if((tab = (GenPolyF2 *)calloc(code->nb_correct, sizeof(GenPolyF2))) == NULL)
	{
		GenPolyF2_free(&s1_inv);
		return 0;
	}
	
	for(i = 0; i < code->nb_correct; ++i)
	{
		if(! GenPolyF2_copy(&sum, syndroms + i + 1))
			goto error_zone;
		for(j = 1; j <= i; ++j)
		{
			if(! GenPolyF2_mul(&tmp, syndroms + j, tab + (i - j), &(code->based_poly)))
			{
				GenPolyF2_free(&sum);
				goto error_zone;
			}
			if(! GenPolyF2_self_add(&sum, &tmp))
			{
				GenPolyF2_free(&sum);
				GenPolyF2_free(&tmp);
				goto error_zone;
			}
			GenPolyF2_free(&tmp);
		}
		if(! GenPolyF2_mul(tab + i, &sum, &s1_inv, &(code->based_poly)))
		{
			GenPolyF2_free(&sum);
			goto error_zone;
		}
		GenPolyF2_free(&sum);
	}
	
	GenPolyF2_free(&s1_inv);
	*racines = tab;
	
	return 1;
	
	error_zone:;
	
	GenPolyF2_free(&s1_inv);
	for(--i; i >= 0; --i)
		GenPolyF2_free(tab + i);
	free(tab);
	
	return 0;
}
#endif

static int chien_calcul(const BCH * code, GenPolyF2 * message, const CPoly * racines, int k)
{
	/*GenPolyF2 poly;
	GenPolyF2 sum;
	GenPolyF2 tmp;
	int i;
	
	if(! GenPolyF2_create_single(&poly, PT_NORMAL, k))
		return 0;
	if(! GenPolyF2_create(&sum, PT_NORMAL, 0))
	{
		GenPolyF2_free(&poly);
		return 0;
	}
	for(i = 0; i < code->nb_correct; ++i)
	{
		if(! GenPolyF2_power(&tmp, &poly, i + 1, &(code->based_poly)))
			goto error_zone;
		if(! GenPolyF2_self_mul(&tmp, racines + i, &(code->based_poly)))
		{
			GenPolyF2_free(&tmp);
			goto error_zone;
		}
		if(! GenPolyF2_self_add(&sum, &tmp))
		{
			GenPolyF2_free(&tmp);
			goto error_zone;
		}
		GenPolyF2_free(&tmp);
	}
	GenPolyF2_free(&poly);
	
	if(GenPolyF2_equals_vals(&sum, 1, -1))
	{
		printf("Correction erreur en %d (%d - %d)\n", code->len - k, code->len, k);
		if(! GenPolyF2_add_val(message, 1, code->len - k))
		{
			GenPolyF2_free(&sum);
			return 0;
		}
	}
	GenPolyF2_free(&sum);*/
	
	GenPolyF2 alpha;
	GenPolyF2 res;
	
	if(! GenPolyF2_create_single(&alpha, PT_NORMAL, k))
		return 0;
	if(! CPoly_eval(&res, racines, &alpha))
	{
		GenPolyF2_free(&alpha);
		return 0;
	}
	GenPolyF2_free(&alpha);
	
	if(GenPolyF2_equals_vals(&res, 0, -1))
	{
		printf("Correction erreur en %d\n", code->len - k);
		if(! GenPolyF2_add_val(message, 1, code->len - k))
		{
			GenPolyF2_free(&res);
			return 0;
		}
	}
	GenPolyF2_free(&res);
	
	return 1;
	
	/*error_zone:;
	
	GenPolyF2_free(&poly);
	GenPolyF2_free(&sum);
	
	return 0;*/
}

static int correct_message(const BCH * code, GenPolyF2 * message, const CPoly * racines)
{
	int i;
	for(i = 0; i < code->len; ++i)
	{
		if(! chien_calcul(code, message, racines, i + 1))
			return 0;
	}
	return 1;
}

static int copy_n(GenPolyF2 * dest, const GenPolyF2 * src, int degree)
{
	int i;
	unsigned char val;
	int src_degree;
	
	src_degree = GenPolyF2_real_degree(src);
	if(degree < src_degree)
		degree = src_degree;
	if(! GenPolyF2_create(dest, PT_NORMAL, degree))
		return 0;
	
	for(i = 0; i <= src_degree; ++i)
	{
		GenPolyF2_get_val(src, &val, i);
		GenPolyF2_set_val(dest, val, i);
	}
	
	return 1;
}

int BCH_is_error(const BCH * code, const GenPolyF2 * message)
{
	GenPolyF2 errors;
	GenPolyF2 tmp;
	if(! GenPolyF2_mul(&tmp, message, &(code->h), NULL))
		return 1;
	if(! GenPolyF2_euclid_div(NULL, &errors, &tmp, &(code->g)))
	{
		GenPolyF2_free(&tmp);
		return 1;
	}
	GenPolyF2_free(&tmp);
	if(GenPolyF2_equals_vals(&errors, 0, -1))
	{
		GenPolyF2_free(&errors);
		return 0;
	}
	GenPolyF2_free(&errors);
	return 1;
}

int BCH_decode(const BCH * code, GenPolyF2 * res, const GenPolyF2 * message)
{
	GenPolyF2 * syndroms = NULL;
	/*GenPolyF2 * racines = NULL;*/
	CPoly racines;
	GenPolyF2 message_copy;
	/*int nb_racines = 0;*/
	
	if(! BCH_is_error(code, message))
	{
		printf("Pas d'erreurs\n");
		if(! GenPolyF2_crush(res, message, code->len - code->msg_len))
			return 0;
	}
	else
		printf("Erreurs à corriger\n");
	
	if(! get_syndroms(code, message, &syndroms, code->nb_correct * 2))
		return 0;
	/*if(! Peterson_solve(&(code->based_poly), syndroms, code->nb_correct, &racines))
	{
		free_gen_tab(&syndroms, code->nb_correct * 2);
		return 0;
	}*/
	if(! get_racines(code, syndroms, &racines))
	{
		free_gen_tab(&syndroms, code->nb_correct * 2);
		return 0;
	}
	/*printf("Localisateur d'erreurs:\n");
	CPoly_print(&racines);*/
	/*printf("Racines:\n");
	print_gen_tab(racines, nb_racines);*/
	/*printf("Syndromes:\n");
	print_gen_tab(syndroms, code->nb_correct * 2);*/
	free_gen_tab(&syndroms, code->nb_correct * 2);
	
	if(! copy_n(&message_copy, message, code->len + 1))
		goto error_zone;
	if(! correct_message(code, &message_copy, &racines))
	{
		GenPolyF2_free(&message_copy);
		goto error_zone;
	}
	if(! GenPolyF2_crush(res, &message_copy, code->len - code->msg_len))
	{
		GenPolyF2_free(&message_copy);
		goto error_zone;
	}
	
	GenPolyF2_free(&message_copy);
	/*free_gen_tab(&racines, nb_racines);*/
	CPoly_free(&racines);
	
	return 1;
	
	error_zone:;
	
	/*free_gen_tab(&racines, nb_racines);*/
	CPoly_free(&racines);
	
	return 0;
}

int Hamming_distance(const GenPolyF2 * poly)
{
	int i;
	int degree;
	unsigned char val;
	int d = 0;
	
	degree = GenPolyF2_real_degree(poly);
	
	for(i = 0; i <= degree; ++i)
	{
		GenPolyF2_get_val(poly, &val, i);
		d += (val != 0);
	}
	return d;
}

int max_correct(int based_poly_degree)
{
	return (1 << (based_poly_degree - 2)) - 1;
}