#include "peterson_solve.h"

#include <stdio.h>
#include <stdlib.h>

int PMatrice_create(PMatrice * matrice, int lignes, int colonnes, const GenPolyF2 * based_poly)
{
	if(! GenPolyF2_copy(&(matrice->based_poly), based_poly))
		return 0;
	if((matrice->seq = (GenPolyF2 *)calloc(lignes * colonnes, sizeof(GenPolyF2))) == NULL)
	{
		GenPolyF2_free(&(matrice->based_poly));
		return 0;
	}
	matrice->lignes = lignes;
	matrice->colonnes = colonnes;
	return 1;
}

void PMatrice_free(PMatrice * matrice)
{
	int i;
	GenPolyF2_free(&(matrice->based_poly));
	if(matrice->seq != NULL)
	{
		for(i = 0; i < matrice->lignes * matrice->colonnes; ++i)
			GenPolyF2_free(matrice->seq + i);
		free(matrice->seq);
		matrice->seq = NULL;
	}
	matrice->lignes = 0;
	matrice->colonnes = 0;
}

GenPolyF2 * PMatrice_val(PMatrice * matrice, int ligne, int colonne)
{
	if(ligne < 0 || colonne < 0 || ligne >= matrice->lignes || colonne >= matrice->colonnes)
		return NULL;
	return matrice->seq + colonne + (ligne * matrice->colonnes);
}

int PMatrice_syndrom_create(PMatrice * matrice, const GenPolyF2 * syndroms, int nb_correct, const GenPolyF2 * based_poly)
{
	int i, j;
	
	if(! PMatrice_create(matrice, nb_correct + 1, nb_correct + 1, based_poly))
		return 0;
	if(! GenPolyF2_create_single(matrice->seq, PT_NORMAL, 0))
	{
		PMatrice_free(matrice);
		return 0;
	}
	for(i = 0; i < matrice->lignes; ++i)
	{
		for(j = 0; j < matrice->colonnes; ++j)
		{
			if(i + j == 0)
				continue;
			if(! GenPolyF2_copy(PMatrice_val(matrice, i, j), syndroms + i + j - 1))
			{
				PMatrice_free(matrice);
				return 0;
			}
		}
	}
	return 1;
}

static void GenPolyF2_simple_print(const GenPolyF2 * poly, int nb)
{
	int degree;
	int i;
	unsigned char val;
	
	degree = GenPolyF2_real_degree(poly);
	for(i = (degree > nb) ? degree : nb; i >= 0; --i)
	{
		if(i > degree)
		{
			putchar(' ');
			continue;
		}
		GenPolyF2_get_val(poly, &val, i);
		putchar('0' + val);
	}
}

void PMatrice_print(PMatrice * matrice)
{
	int base_degree;
	int i, j;
	
	base_degree = GenPolyF2_real_degree(&(matrice->based_poly));
	for(i = 0; i < matrice->lignes; ++i)
	{
		for(j = 0; j < matrice->colonnes; ++j)
		{
			GenPolyF2_simple_print(PMatrice_val(matrice, i, j), base_degree - 1);
			putchar(' ');
		}
		putchar('\n');
	}
}

int PMatrice_mul_line(PMatrice * matrice, int ligne, const GenPolyF2 * poly)
{
	int i;
	for(i = 0; i < matrice->colonnes; ++i)
	{
		if(! GenPolyF2_self_mul(PMatrice_val(matrice, ligne, i), poly, &(matrice->based_poly)))
			return 0;
	}
	return 1;
}

int PMatrice_add_lines(PMatrice * matrice, int ligne_un, int ligne_deux)
{
	int i;
	for(i = 0; i < matrice->colonnes; ++i)
	{
		if(! GenPolyF2_self_add(PMatrice_val(matrice, ligne_un, i), PMatrice_val(matrice, ligne_deux, i)))
			return 0;
	}
	return 1;
}

int PMatrice_swap_lines(PMatrice * matrice, int ligne_un, int ligne_deux)
{
	GenPolyF2 tmp;
	int i;
	for(i = 0; i < matrice->colonnes; ++i)
	{
		tmp = *(PMatrice_val(matrice, ligne_un, i));
		*(PMatrice_val(matrice, ligne_un, i)) = *(PMatrice_val(matrice, ligne_deux, i));
		*(PMatrice_val(matrice, ligne_deux, i)) = tmp;
	}
	return 1;
}

int PMatrice_reduce(PMatrice * matrice, int id)
{
	int i;
	GenPolyF2 inv;
	
	for(i = matrice->lignes - 1 - id; i >= 0; --i)
	{
		if(! GenPolyF2_invert(&inv, PMatrice_val(matrice, i, id), &(matrice->based_poly)))
			continue;
		if(! PMatrice_mul_line(matrice, i, &inv))
		{
			GenPolyF2_free(&inv);
			return 0;
		}
		GenPolyF2_free(&inv);
	}
	for(i = matrice->lignes - id - 1; i >= 0; --i)
	{
		if(! GenPolyF2_equals_vals(PMatrice_val(matrice, i, id), 0, -1))
			continue;
		PMatrice_swap_lines(matrice, i, 0);
	}
	for(i = matrice->lignes - id - 2; i >= 0; --i)
	{
		if(! PMatrice_add_lines(matrice, i, matrice->lignes - id - 1))
			return 0;
	}
	return 1;
}

int PMatrice_solve(PMatrice * matrice)
{
	int i;
	for(i = 0; i < matrice->lignes - 1; ++i)
	{
		/*PMatrice_print(matrice);*/
		if(! PMatrice_reduce(matrice, i))
			return 0;
		/*printf("etape %d\n", i);*/
	}
	/*PMatrice_print(matrice);*/
	return 1;
}

int PMatrice_check(PMatrice * matrice)
{
	int i;
	for(i = 0; i < matrice->lignes - 1; ++i)
	{
		if(GenPolyF2_equals_vals(PMatrice_val(matrice, matrice->lignes - 1 - i, i), 0, -1))
			return 0;
	}
	return 1;
}

int Peterson_solve(const GenPolyF2 * based_poly, const GenPolyF2 * syndroms, int nb_correct, GenPolyF2 ** racines)
{
	PMatrice mat;
	GenPolyF2 * tab = NULL;
	GenPolyF2 tmp;
	int i, k;
	int nb_racines;

	nb_racines = nb_correct;
	if(! PMatrice_syndrom_create(&mat, syndroms, nb_correct, based_poly))
		return 0;
	if(! PMatrice_solve(&mat))
	{
		PMatrice_free(&mat);
		return 0;
	}
	/*printf("check: %d\n", PMatrice_check(&mat));*/
	while(! PMatrice_check(&mat))
	{
		--nb_racines;
		PMatrice_free(&mat);
		if(nb_racines <= 0)
			return 0;
		if(! PMatrice_syndrom_create(&mat, syndroms, nb_racines, based_poly))
			return 0;
		if(! PMatrice_solve(&mat))
		{
			PMatrice_free(&mat);
			return 0;
		}
	}
	if((tab = (GenPolyF2 *)calloc(nb_correct, sizeof(GenPolyF2))) == NULL)
	{
		PMatrice_free(&mat);
		return 0;
	}
	for(i = 0; i < nb_racines; ++i)
	{
		if(! GenPolyF2_copy(tab + i, PMatrice_val(&mat, i + 1, mat.colonnes - 1)))
			goto error_zone;
		for(k = 0; k < i; ++k)
		{
			if(! GenPolyF2_mul(&tmp, tab + k, PMatrice_val(&mat, i + 1, mat.colonnes - 2 - k), based_poly))
				goto error_zone;
			if(! GenPolyF2_self_add(tab + i, &tmp))
			{
				GenPolyF2_free(&tmp);
				goto error_zone;
			}
			GenPolyF2_free(&tmp);
		}
	}
	for(i = nb_racines; i < nb_correct; ++i)
	{
		if(! GenPolyF2_create(tab + i, PT_NORMAL, 0))
			goto error_zone;
	}
	
	PMatrice_free(&mat);
	
	*racines = tab;
	
	return 1;
	
	error_zone:;
	
	free_gen_tab(&tab, nb_correct);
	PMatrice_free(&mat);
	
	return 0;
}

#include <SDL/SDL.h>

int Gauss_solve(const GenPolyF2 * based_poly, const GenPolyF2 * syndroms, int nb_correct, GenPolyF2 ** solutions, int * nb_solutions)
{
	PMatrice mat;
	GenPolyF2 tmp;
	int i;
	int j;
	if(! PMatrice_syndrom_create(&mat, syndroms, nb_correct, based_poly))
		return 0;
	for(i = 0; i <= nb_correct; ++i)
	{
		if(! GenPolyF2_create(&tmp, PT_NORMAL, 0))
		{
			PMatrice_free(&mat);
			return 0;
		}
		GenPolyF2_free(PMatrice_val(&mat, nb_correct, i));
		*(PMatrice_val(&mat, nb_correct, i)) = tmp;
	}
	/*printf("initial\n");
	PMatrice_print(&mat);*/
	for(i = 0; i < nb_correct; ++i)
	{
		if(GenPolyF2_equals_vals(PMatrice_val(&mat, i, i), 0, -1))
		{
			for(j = i + 1; j < nb_correct; ++j)
			{
				if(j == i)
					continue;
				if(! GenPolyF2_equals_vals(PMatrice_val(&mat, j, i), 0, -1))
					PMatrice_swap_lines(&mat, i, j);
			}
			if(GenPolyF2_equals_vals(PMatrice_val(&mat, i, i), 0, -1))
				goto no_solution;
		}
		GenPolyF2_invert(&tmp, PMatrice_val(&mat, i, i), based_poly);
		PMatrice_mul_line(&mat, i, &tmp);
		GenPolyF2_free(&tmp);
		for(j = 0; j < nb_correct; ++j)
		{
			if(i == j)
				continue;
			if(GenPolyF2_equals_vals(PMatrice_val(&mat, j, i), 0, -1))
				continue;
			GenPolyF2_invert(&tmp, PMatrice_val(&mat, j, i), based_poly);
			PMatrice_mul_line(&mat, j, &tmp);
			GenPolyF2_free(&tmp);
			PMatrice_add_lines(&mat, j, i);
			/*printf("do %d <- %d\n", j, i);
			PMatrice_print(&mat);*/
		}
	}
	for(i = 0; i < nb_correct; ++i)
	{
		if(GenPolyF2_equals_vals(PMatrice_val(&mat, i, i), 0, -1))
			goto no_solution;
		GenPolyF2_invert(&tmp, PMatrice_val(&mat, i, i), based_poly);
		PMatrice_mul_line(&mat, i, &tmp);
		GenPolyF2_free(&tmp);
		
	}
	/*printf("final\n");
	PMatrice_print(&mat);*/
	if(nb_solutions != NULL)
		*nb_solutions = nb_correct;
	if((*solutions = (GenPolyF2 *)calloc(nb_correct, sizeof(GenPolyF2))) == NULL)
	{
		PMatrice_free(&mat);
		return 0;
	}
	for(i = 0; i < nb_correct; ++i)
	{
		if(! GenPolyF2_copy(*solutions + i, PMatrice_val(&mat, i, nb_correct)))
		{
			PMatrice_free(&mat);
			for(--i; i >= 0; --i)
				GenPolyF2_free(*solutions + i);
			free(*solutions);
			*solutions = NULL;
			return 0;
		}
	}
	PMatrice_free(&mat);
	return 1;
	
	no_solution:;
	/*PMatrice_print(&mat);*/
	PMatrice_free(&mat);
	
	return Gauss_solve(based_poly, syndroms, nb_correct - 1, solutions, nb_solutions);
}