#include "poly.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

int size_degree(int degree)
{
	if(degree < 0)
		return 0;
	++degree;
	return (degree / 8) + (degree % 8 != 0);
}

static int degree_id(int degree)
{
	return size_degree(degree) - 1;
}

static unsigned char * get_seq_octet(const PolyF2 * poly, int id)
{
	int did;
	did = degree_id(id);
	if(did < 0 || id > poly->degree)
		return NULL;
	return poly->bit_seq + did;
}

static int max(int a, int b)
{
	return (a > b) ? a : b;
}

/* Init functions */

int PolyF2_create(PolyF2 * poly, int degree)
{
	void * bit_seq = NULL;
	
	if((bit_seq = (unsigned char *)calloc(size_degree(degree), sizeof(unsigned char))) == NULL)
		return 0;
	poly->bit_seq = bit_seq;
	poly->degree = degree;
	
	return 1;
}

int PolyF2_create_by_vals_aux(PolyF2 * poly, int i, va_list ap)
{
	int val;
	if((val = va_arg(ap, int)) < 0)
		return PolyF2_create(poly, i - 1);
	if(! PolyF2_create_by_vals_aux(poly, i + 1, ap))
		return 0;
	if(! PolyF2_set_val(poly, val, i))
	{
		if(i == 0)
			PolyF2_free(poly);
		return 0;
	}
	return 1;
}

int PolyF2_create_single(PolyF2 * poly, int degree)
{
	PolyF2 res;
	if(! PolyF2_create(&res, degree))
		return 0;
	if(! PolyF2_set_val(&res, 1, degree))
	{
		PolyF2_free(&res);
		return 0;
	}
	*poly = res;
	return 1;
}

int PolyF2_create_by_vals(PolyF2 * poly, ...)
{
	int success;
	va_list ap;
	va_start(ap, poly);
	success = PolyF2_create_by_vals_va(poly, ap);
	va_end(ap);
	return success;
}

int PolyF2_create_by_vals_va(PolyF2 * poly, va_list ap)
{
	return PolyF2_create_by_vals_aux(poly, 0, ap);
}

void PolyF2_free(PolyF2 * poly)
{
	if(poly == NULL)
		return;
	if(poly->bit_seq != NULL)
	{
		free(poly->bit_seq);
		poly->bit_seq = NULL;
	}
	poly->degree = 0;
}

int PolyF2_set_val(PolyF2 * poly, unsigned char val, int id)
{
	unsigned char * ptr = NULL;
	if((ptr = get_seq_octet(poly, id)) == NULL)
		return 0;
	id %= 8;
	*ptr &= ~((unsigned char)1 << id);
	*ptr |= ((unsigned char)(val != 0) << id);
	return 1;
}

int PolyF2_set_vals(PolyF2 * poly, ...)
{
	int success;
	va_list ap;
	
	va_start(ap, poly);
	success = PolyF2_set_vals_va(poly, ap);
	va_end(ap);
	
	return success;
}

int PolyF2_set_vals_va(PolyF2 * poly, va_list ap)
{
	int i;
	int val;
	for(i = 0; i <= poly->degree && (val = va_arg(ap, int)) >= 0; ++i)
	{
		if(! PolyF2_set_val(poly, val != 0, i))
			return 0;
	}
	return 1;
}

int PolyF2_get_val(const PolyF2 * poly, unsigned char * val, int id)
{
	unsigned char * seq = NULL;
	if((seq = get_seq_octet(poly, id)) == NULL)
		return 0;
	if(val != NULL)
		*val = (((*seq) & ((unsigned char)1 << (id % 8))) != 0);
	return 1;
}

int PolyF2_add_val(PolyF2 * poly, unsigned char val, int id)
{
	unsigned char * ptr = NULL;
	if((ptr = get_seq_octet(poly, id)) == NULL)
		return 0;
	id %= 8;
	*ptr ^= ((unsigned char)(val != 0) << id);
	return 1;
}

int PolyF2_copy(PolyF2 * dest, const PolyF2 * src)
{
	int size;
	int i;
	PolyF2 res;
	
	if((size = size_degree(src->degree)) <= 0)
		return 0;
	if(! PolyF2_create(&res, src->degree))
		return 0;
	for(i = 0; i < size; ++i)
		*(res.bit_seq + i) = *(src->bit_seq + i);
	
	*dest = res;
	return 1;
}

/* Operation functions */

int PolyF2_add(PolyF2 * res, const PolyF2 * first, const PolyF2 * second)
{
	PolyF2 _res;
	int degree;
	int size_degree_val;
	int first_degree;
	int second_degree;
	int first_size_degree;
	int second_size_degree;
	int i;
	
	first_degree = PolyF2_real_degree(first);
	second_degree = PolyF2_real_degree(second);
	if((degree = max(first_degree, second_degree)) < 0)
		return 0;
	first_size_degree = size_degree(first_degree);
	second_size_degree = size_degree(second_degree);
	size_degree_val = max(first_size_degree, second_size_degree);
	
	if(! PolyF2_create(&_res, degree))
		return 0;
		
	for(i = 0; i < size_degree_val; ++i)
	{
		if((i < first_size_degree) ^ (i < second_size_degree))
		{
			if(i < first_size_degree)
				*(_res.bit_seq + i) = *(first->bit_seq + i);
			else
				*(_res.bit_seq + i) = *(second->bit_seq + i);
		}
		else
			*(_res.bit_seq + i) = *(first->bit_seq + i) ^ *(second->bit_seq + i);
	}
	
	*res = _res;
	return 1;
}

int PolyF2_sub(PolyF2 * res, const PolyF2 * first, const PolyF2 * second)
{
	/* Same as add here */
	return PolyF2_add(res, first, second);
}

int PolyF2_mul(PolyF2 * res, const PolyF2 * first, const PolyF2 * second, const PolyF2 * based_poly)
{
	PolyF2 _res;
	int degree;
	int first_degree;
	int second_degree;
	int i;
	int j;
	unsigned char val;
	unsigned char val_aux;
	int success = 1;
	
	first_degree = PolyF2_real_degree(first);
	second_degree = PolyF2_real_degree(second);
	if((degree = first_degree + second_degree) < 0)
		return 0;
	if(! PolyF2_create(&_res, degree))
		return 0;
	for(i = 0; i <= second_degree; ++i)
	{
		if(! PolyF2_get_val(second, &val, i))
			goto error_zone;
		if(! val)
			continue;
		for(j = 0; j <= first_degree; ++j)
		{
			if(! PolyF2_get_val(first, &val_aux, j))
				goto error_zone;
			if(! val_aux)
				continue;
			if(! PolyF2_add_val(&_res, val_aux, j + i))
				goto error_zone;
		}
	}
	if(based_poly == NULL)
	{
		*res = _res;
		return 1;
	}
	success = PolyF2_euclid_div(NULL, res, &_res, based_poly);
	PolyF2_free(&_res);
	return success;
	
	error_zone:;
	
	PolyF2_free(&_res);
	return 0;
}

int PolyF2_mul_degree(PolyF2 * poly, int degree)
{
	PolyF2 res;
	int poly_deg;
	int i;
	unsigned char val;
	if(degree < 0)
		return 0;
	poly_deg = PolyF2_real_degree(poly);
	if(! PolyF2_create(&res, poly_deg + degree))
		return 0;
	for(i = poly_deg; i >= 0; --i)
	{
		if(! PolyF2_get_val(poly, &val, i))
			goto error_zone;
		if(! PolyF2_set_val(&res, val, i + degree))
			goto error_zone;
	}
	PolyF2_free(poly);
	*poly = res;
	return 1;
	
	error_zone:;
	
	PolyF2_free(&res);
	return 0;
}

static int PolyF2_euclid_div_aux(PolyF2 * factor, PolyF2 * remain, PolyF2 * poly, PolyF2 * divisor, int max_deg)
{
	PolyF2 poly_mul;
	PolyF2 new_poly;
	int poly_deg;
	int div_deg;
	poly_deg = PolyF2_real_degree(poly);
	div_deg = PolyF2_real_degree(divisor);
	
	if(div_deg == 0)
	{
		PolyF2 _remain;
		unsigned char val;
		if(! PolyF2_get_val(divisor, &val, 0))
			return 0;
		if(val == 0)
			return 0;
		if(! PolyF2_create_by_vals(&_remain, 0, -1))
			return 0;
		if(! PolyF2_copy(factor, poly))
		{
			PolyF2_free(&_remain);
			return 0;
		}
		*remain = _remain;
		return 1;
	}
	if(poly_deg < div_deg)
	{
		PolyF2 _remain;
		if(! PolyF2_copy(&_remain, poly))
			return 0;
		if(! PolyF2_create(factor, max_deg))
		{
			PolyF2_free(&_remain);
			return 0;
		}
		*remain = _remain;
		return 1;
	}
	if(! PolyF2_copy(&poly_mul, divisor))
		return 0;
	if(! PolyF2_mul_degree(&poly_mul, poly_deg - div_deg))
		goto error_zone;
	if(! PolyF2_sub(&new_poly, poly, &poly_mul))
		goto error_zone;
	PolyF2_free(poly);
	*poly = new_poly;
	PolyF2_free(&poly_mul);
	
	if(! PolyF2_euclid_div_aux(factor, remain, poly, divisor, max(poly_deg - div_deg, max_deg)))
		return 0;
	if(! PolyF2_set_val(factor, 1, poly_deg - div_deg))
		return 0;
	return 1;
	
	error_zone:;
	
	PolyF2_free(&poly_mul);
	return 0;
}

int PolyF2_euclid_div(PolyF2 * factor, PolyF2 * remain, const PolyF2 * poly, const PolyF2 * divisor)
{
	PolyF2 poly_copy, div_copy, _factor, _remain;
	int success;
	
	if(! PolyF2_copy(&poly_copy, poly))
		return 0;
	if(! PolyF2_copy(&div_copy, divisor))
	{
		PolyF2_free(&poly_copy);
		return 0;
	}
	success = PolyF2_euclid_div_aux(&_factor, &_remain, &poly_copy, &div_copy, 1);
	
	if(factor != NULL)
		*factor = _factor;
	else
		PolyF2_free(&_factor);
	if(remain != NULL)
		*remain = _remain;
	else
		PolyF2_free(&_remain);
		
	PolyF2_free(&poly_copy);
	PolyF2_free(&div_copy);
	return success;
}

int PolyF2_equals(const PolyF2 * first, const PolyF2 * second)
{
	int degree;
	int size;
	int i;
	
	if((degree = PolyF2_real_degree(first)) != PolyF2_real_degree(second))
		return 0;
	size = size_degree(degree);
	for(i = 0; i < size; ++i)
	{
		if(*(first->bit_seq + i) != *(second->bit_seq + i))
			return 0;
	}
	return 1;
}

/* Information functions */

int PolyF2_degree(const PolyF2 * poly)
{
	return poly->degree;
}

int PolyF2_real_degree(const PolyF2 * poly)
{
	int i;
	unsigned char val;
	for(i = poly->degree; i >= 0; --i)
	{
		if(! PolyF2_get_val(poly, &val, i))
			return -1;
		if(val)
			return i;
	}
	return 0;
}

/* In/Out functions */

void PolyF2_print(const PolyF2 * poly)
{
	int i;
	int deg;
	unsigned char first_coeff = 1;
	unsigned char * ptr = NULL;
	deg = PolyF2_real_degree(poly);
	for(i = 0; i <= deg; ++i)
	{
		if(i % 8 == 0 && (ptr = get_seq_octet(poly, i)) == NULL)
			return;
		if(*ptr & (1 << (i % 8)))
		{
			if(! first_coeff)
				printf(" + ");
			else
				first_coeff = 0;
			if(i == 0)
				printf("1");
			else
			{
				printf("x");
				if(i > 1)
					printf("^%d", i);
			}
		}
		else if(i == 0 && deg <= 0)
			printf("0");
	}
	printf("\n");
}

void PolyF2_print_bin(const PolyF2 * poly)
{
	int i;
	int deg;
	unsigned char val;
	deg = PolyF2_real_degree(poly);
	#ifdef DEBUG
	printf("d : %d/%d | ", deg, poly->degree);
	#endif
	for(i = deg; i >= 0; --i)
	{
		if(! PolyF2_get_val(poly, &val, i))
			return;
		putchar('0' + val);
	}
	putchar('\n');
}