#include "poly_int.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/* Init functions */

int PolyIntF2_create(PolyIntF2 * poly, int degree)
{
	poly->bit_seq = 0;
	poly->degree = degree;
	
	return 1;
}

int PolyIntF2_create_by_vals(PolyIntF2 * poly, ...)
{
	va_list ap;
	int success;
	
	va_start(ap, poly);
	success = PolyIntF2_create_by_vals_va(poly, ap);
	va_end(ap);
	
	return success;
}

int PolyIntF2_create_by_vals_va(PolyIntF2 * poly, va_list ap)
{
	int i;
	int val;
	
	poly->bit_seq = 0;
	for(i = 0; (val = va_arg(ap, int)) >= 0; ++i)
	{
		if(! PolyIntF2_set_val(poly, val, i))
			return 0;
	}
	poly->degree = PolyIntF2_real_degree(poly);
	return 1;
}

int PolyIntF2_create_single(PolyIntF2 * poly, int degree)
{
	poly->bit_seq = (BS_TYPE)1 << degree;
	poly->degree = degree;
	
	return 1;
}

void PolyIntF2_free(PolyIntF2 * poly)
{
	poly->bit_seq = 0;
	poly->degree = 0;
}

int PolyIntF2_set_val(PolyIntF2 * poly, unsigned char val, int id)
{
	poly->bit_seq &= ~((BS_TYPE)1 << id);
	poly->bit_seq |= ((BS_TYPE)(val)) << id;
	return 1;
}

int PolyIntF2_set_vals(PolyIntF2 * poly, ...)
{
	va_list ap;
	int success;
	
	va_start(ap, poly);
	success = PolyIntF2_set_vals_va(poly, ap);
	va_end(ap);
	
	return success;
}

int PolyIntF2_set_vals_va(PolyIntF2 * poly, va_list ap)
{
	int i;
	int val;
	
	for(i = 0; (val = va_arg(ap, int)) >= 0 && i <= poly->degree; ++i)
	{
		if(! PolyIntF2_set_val(poly, val != 0, i))
			return 0;
	}
	
	return 1;
}

int PolyIntF2_get_val(const PolyIntF2 * poly, unsigned char * val, int id)
{
	if(val != NULL)
		*val = ((poly->bit_seq) & ((BS_TYPE)1 << id)) != 0;
	return 1;
}

int PolyIntF2_add_val(PolyIntF2 * poly, unsigned char val, int id)
{
	if(id > poly->degree)
		return 0;
	poly->bit_seq ^= ((BS_TYPE)(val) << id);
	return 1;
}

int PolyIntF2_copy(PolyIntF2 * dest, const PolyIntF2 * src)
{
	dest->bit_seq = src->bit_seq;
	dest->degree = PolyIntF2_real_degree(src);
	return 1;
}

/* Operation functions */

int PolyIntF2_mul_degree(PolyIntF2 * poly, int degree)
{
	(poly->bit_seq) <<= degree;
	(poly->degree) += degree;
	return 1;
}

int PolyIntF2_add(PolyIntF2 * res, const PolyIntF2 * first, const PolyIntF2 * second)
{
	res->bit_seq = (first->bit_seq) ^ (second->bit_seq);
	res->degree = PolyIntF2_real_degree(res);
	return 1;
}

int PolyIntF2_sub(PolyIntF2 * res, const PolyIntF2 * first, const PolyIntF2 * second)
{
	return PolyIntF2_add(res, first, second);
}

int PolyIntF2_mul(PolyIntF2 * res, const PolyIntF2 * first, const PolyIntF2 * second, const PolyIntF2 * based_poly)
{
	int i;
	res->bit_seq = 0;
	res->degree = PolyIntF2_real_degree(first) + PolyIntF2_real_degree(second);
	for(i = 0; i <= second->degree; ++i)
		if(second->bit_seq & ((BS_TYPE)1 << i))
			(res->bit_seq) ^= (first->bit_seq) << i;
	if(based_poly != NULL)
	{
		PolyIntF2 remain;
		if(! PolyIntF2_euclid_div(NULL, &remain, res, based_poly))
			return 0;
		*res = remain;
	}
	return 1;
}

int PolyIntF2_euclid_div(PolyIntF2 * factor, PolyIntF2 * remain, const PolyIntF2 * poly, const PolyIntF2 * divisor)
{
	PolyIntF2 _remain;
	PolyIntF2 _factor;
	int div_degree;
	int rem_degree;
	
	div_degree = PolyIntF2_real_degree(divisor);
	if(! PolyIntF2_copy(&_remain, poly))
		return 0;
	rem_degree = PolyIntF2_real_degree(&_remain);
	_factor.bit_seq = 0;
	while(div_degree <= rem_degree)
	{
		(_remain.bit_seq) ^= (divisor->bit_seq) << (rem_degree - div_degree);
		(_factor.bit_seq) |= (BS_TYPE)1 << (rem_degree - div_degree);
		rem_degree = PolyIntF2_real_degree(&_remain);
	}
	_remain.degree = rem_degree;
	_factor.degree = PolyIntF2_real_degree(&_factor);
	if(remain != NULL)
		*remain = _remain;
	if(factor != NULL)
		*factor = _factor;
	return 1;
}

int PolyIntF2_equals(const PolyIntF2 * first, const PolyIntF2 * second)
{
	return first->bit_seq == second->bit_seq;
}

/* Information functions */

int PolyIntF2_degree(const PolyIntF2 * poly)
{
	return poly->degree;
}

int PolyIntF2_real_degree(const PolyIntF2 * poly)
{
	int i;
	BS_TYPE seq;
	if(poly->bit_seq == 0)
		return 0;
	for(i = 0, seq = poly->bit_seq; seq > 0; seq >>= 1, ++i);
	return i - 1;
}

/* In/Out functions */

void PolyIntF2_print(const PolyIntF2 * poly)
{
	int i;
	int deg;
	unsigned char val;
	int first_coeff = 1;
	
	if(poly->bit_seq == 0)
	{
		printf("0\n");
		return;
	}
	deg = PolyIntF2_real_degree(poly);
	for(i = deg; i >= 0; --i)
	{
		if(! PolyIntF2_get_val(poly, &val, i))
			return;
		if(val)
		{
			if(! first_coeff)
				printf(" + ");
			else
				first_coeff = 0;
			if(i == 0)
				printf("1");
			else
			{
				printf("x");
				if(i > 1)
					printf("^%d", i);
			}
		}
	}
	printf("\n");
}

void PolyIntF2_print_bin(const PolyIntF2 * poly)
{
	int i;
	int deg;
	unsigned char val;
	
	deg = PolyIntF2_real_degree(poly);
	#ifdef DEBUG
	printf("d : %d/%d | ", deg, poly->degree);
	#endif
	for(i = deg; i >= 0; --i)
	{
		if(! PolyIntF2_get_val(poly, &val, i))
			continue;
		putchar('0' + val);
	}
	putchar('\n');
}