#include "qr_code.h"

#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "generic_poly.h"
#include "bch.h"
#include "parse_poly.h"

/* Global init */

static SDL_Surface * screen = NULL;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	static Uint32 rmask = 0xff000000;
	static Uint32 gmask = 0x00ff0000;
	static Uint32 bmask = 0x0000ff00;
	static Uint32 amask = 0x000000ff;
	
	int r_shift = 24;
	int g_shift = 16;
	int b_shift = 8;
	int a_shift = 0;
#else
	static Uint32 rmask = 0x000000ff;
	static Uint32 gmask = 0x0000ff00;
	static Uint32 bmask = 0x00ff0000;
	static Uint32 amask = 0xff000000;
	
	int r_shift = 0;
	int g_shift = 8;
	int b_shift = 16;
	int a_shift = 24;
#endif

#ifdef ANIM
static int at_end = 1;
#endif

static void SDL_Surface_get_pixel(const SDL_Surface * surface, int x, int y, int * r, int * g, int * b, int * a)
{
	Uint32 color;
	color = *((Uint32 *)(surface->pixels) + ((y * surface->pitch) / 4) + x);
	if(r != NULL)
		*r = ((Uint32)(color & rmask) >> r_shift);
	if(g != NULL)
		*g = ((Uint32)(color & gmask) >> g_shift);
	if(b != NULL)
		*b = ((Uint32)(color & bmask) >> b_shift);
	if(a != NULL)
		*a = ((Uint32)(color & amask) >> a_shift);
}

static void SDL_Surface_set_pixel(SDL_Surface * surface, int x, int y, int r, int g, int b, int a)
{
	Uint32 color;
	color = (((Uint32)r << r_shift) & rmask)
		| (((Uint32)g << g_shift) & gmask)
		| (((Uint32)b << b_shift) & bmask)
		| (((Uint32)a << a_shift) & amask);
	SDL_LockSurface(surface);
	*((Uint32 *)(surface->pixels) + ((y * surface->pitch) / 4) + x) = color;
	SDL_UnlockSurface(surface);
}

static int SDL_Surface_is_free_bit(SDL_Surface * surface, int x, int y)
{
	int r, g, b, a;
	SDL_Surface_get_pixel(surface, x, y, &r, &g, &b, &a);
	return r == 127 && g == 127 && b == 127;
}

static int SDL_Surface_get_bit(SDL_Surface * surface, int x, int y)
{
	int r, g, b, a;
	SDL_Surface_get_pixel(surface, x, y, &r, &g, &b, &a);
	return (r + g + b) < 127 * 3;
}

static void SDL_Surface_set_bit(SDL_Surface * surface, int x, int y, int bit)
{
	bit = (bit > 0) ? 0 : 255;
	SDL_Surface_set_pixel(surface, x, y, bit, bit, bit, 255);
}

static SDL_Rect get_rect(int x, int y, int w, int h)
{
	SDL_Rect rect;
	rect.x = x;
	rect.y = y;
	rect.w = w;
	rect.h = h;
	return rect;
}

#ifdef ANIM
void print_area(QRCode * code)
{
	SDL_Surface * tmp = NULL;
	if(code->area == NULL)
		return;
	tmp = code->surface;
	code->surface = code->area;
	code->area = tmp;
	QRCode_print_autozoom(code);
	/*QRCode_print_zoom(code, 20 - code->version / 2);*/
	tmp = code->surface;
	code->surface = code->area;
	code->area = tmp;
	/*SDL_SaveBMP(screen, "media/area.bmp");*/
}
#endif

int QRCode_init()
{
	if(SDL_Init(SDL_INIT_VIDEO) == -1)
		return 0;
	screen = NULL;
	return 1;
}

void QRCode_quit()
{
	if(screen != NULL)
		SDL_FreeSurface(screen);
	SDL_Quit();
}

/* Struct init */

int QRCode_create(QRCode * code, int size)
{
	memset(code, 0, sizeof(QRCode));
	if((code->surface = SDL_CreateRGBSurface(SDL_HWSURFACE, size, size, 32,
		rmask, gmask, bmask, amask)) == NULL)
	{
		fprintf(stderr, "Erreur SDL_CreateRGBSurface : %s\n", SDL_GetError());
		return 0;
	}
	code->area = NULL;
	SDL_LockSurface(code->surface);
	SDL_FillRect(code->surface, NULL, SDL_MapRGB(code->surface->format, 127, 127, 127));
	SDL_UnlockSurface(code->surface);
	memset(&(code->current_coord), 0, sizeof(SDL_Rect));
	code->current_id = 0;
	
	code->cl = M;
	code->mask_id = 0;
	code->version = 0;
	code->nb_codewords = 0;
	return 1;
}

static int SDL_Surface_make_clean(SDL_Surface ** surface)
{
	SDL_Surface * tmp = NULL;
	SDL_Rect pos;
	
	if((tmp = SDL_CreateRGBSurface(SDL_HWSURFACE, (*surface)->w, (*surface)->h, 32, rmask, gmask, bmask, amask)) == NULL)
		return 0;
	pos = get_rect(0, 0, (*surface)->w, (*surface)->h);
	SDL_BlitSurface(*surface, NULL, tmp, &pos);
	SDL_FreeSurface(*surface);
	*surface = tmp;
	return 1;
}

int QRCode_open(QRCode * code, const char * path)
{
	SDL_Surface * surface = NULL;
	
	memset(code, 0, sizeof(QRCode));
	if((surface = IMG_Load(path)) == NULL)
		return 0;
	if(! SDL_Surface_make_clean(&surface))
	{
		SDL_FreeSurface(surface);
		return 0;
	}
	if(! Image_to_matrix(surface, &(code->surface)))
	{
		fprintf(stderr, "Erreur de conversion de l'image en image QR code\n");
		SDL_FreeSurface(surface);
		return 0;
	}
	SDL_FreeSurface(surface);
	
	if((code->surface->w - 17) % 4 != 0)
	{
		fprintf(stderr, "Erreur de récupération de la version\n");
		SDL_FreeSurface(code->surface);
		return 0;
	}
	code->version = (code->surface->w - 17) / 4;
	
	if((code->area = SDL_CreateRGBSurface(SDL_HWSURFACE, code->surface->w, code->surface->h, 32, rmask, gmask, bmask, amask)) == NULL)
	{
		SDL_FreeSurface(code->surface);
		return 0;
	}
	SDL_FillRect(code->area, NULL, SDL_MapRGB(code->area->format, 127, 127, 127));
	
	code->current_id = 0;
	
	return 1;
}

void QRCode_free(QRCode * code)
{
	if(code->surface != NULL)
	{
		SDL_FreeSurface(code->surface);
		code->surface = NULL;
	}
	if(code->area != NULL)
	{
		SDL_FreeSurface(code->area);
		code->area = NULL;
	}
	memset(&(code->current_coord), 0, sizeof(SDL_Rect));
	code->current_id = 0;
	if(code->nb_codewords > 0)
		RS_free(&(code->rs));
	code->nb_codewords = 0;
}

int QRCode_set_move(QRCode * code, int (*get_next_coord)(SDL_Rect *, QRCode *, void *),
	int (*get_coord_by_id)(int id, SDL_Rect *, QRCode *, void *),
	int (*is_message_end)(QRCode *, void *), int (*init_data)(QRCode *, void *),
	void * user_data)
{
	if(!((get_next_coord != NULL) ^ (get_coord_by_id != NULL)))
		return 0;
	code->get_next_coord = get_next_coord;
	code->get_coord_by_id = get_coord_by_id;
	code->is_message_end = is_message_end;
	code->user_data_move = user_data;
	if(init_data != NULL)
	{
		if(! (*init_data)(code, user_data))
			return 0;
	}
	if(get_coord_by_id != NULL)
		(*get_coord_by_id)(code->current_id, &(code->current_coord), code, user_data);
	return 1;
}

int QRCode_go_to_next(QRCode * code)
{
	if(code->is_message_end != NULL)
	{
		if((*(code->is_message_end))(code, code->user_data_move))
			return 0;
	}
	if(code->get_next_coord != NULL)
	{
		++(code->current_id);
		return (*(code->get_next_coord))(&(code->current_coord), code, code->user_data_move);
	}
	return (*(code->get_coord_by_id))(++(code->current_id), &(code->current_coord), code, code->user_data_move);
}

int QRCode_read_bit(QRCode * code, unsigned char * bit)
{
	if(code->is_message_end != NULL)
	{
		if((*(code->is_message_end))(code, code->user_data_move))
			return 0;
	}
	if(bit != NULL)
		*bit = SDL_Surface_get_bit(code->surface, code->current_coord.x, code->current_coord.y);
	if(code->area != NULL)
	{
		SDL_Surface_set_pixel(code->area, code->current_coord.x, code->current_coord.y, ((code->current_id / 8) % 2) * 127 + 63, ((code->current_id / 8) % 2) * 127 + 63, ((code->current_id / 8) % 2) * 127 + 63, 255);
	}
	QRCode_go_to_next(code);
	return 1;
}

int QRCode_write_bit(QRCode * code, unsigned char bit)
{
	if(code->is_message_end != NULL)
	{
		if((*(code->is_message_end))(code, code->user_data_move))
			return 0;
	}
	SDL_Surface_set_bit(code->surface, code->current_coord.x, code->current_coord.y, bit);
	QRCode_go_to_next(code);
	return 1;
}

int QRCode_read_data(QRCode * code, void * data, int bit_size,
	int (*set_data)(void * data, int current_id, unsigned char bit))
{
	int i;
	unsigned char bit;
	for(i = 0; i < bit_size; ++i)
	{
		if(! QRCode_read_bit(code, &bit))
			return i;
		if(! (*set_data)(data, i, bit))
			return i;
		/*SDL_SaveBMP(code->area, "media/tmpi.bmp");*/
		#ifdef ANIM
		print_area(code);
		SDL_Delay(25);
		QRCode_stop_print();
		#endif
	}
	return bit_size;
}

int QRCode_write_data(QRCode * code, void * data, int bit_size,
	int (*get_data)(void * data, int current_id, unsigned char * bit))
{
	int i;
	unsigned char bit;
	for(i = 0; i < bit_size; ++i)
	{
		if(! (*get_data)(data, i, &bit))
			return i;
		if(! QRCode_write_bit(code, bit))
			return i;
		#ifdef ANIM
		if(! at_end)
		{
			QRCode_print_autozoom(code);
			/*QRCode_print_zoom(code, 20);*/
			SDL_Delay(25);
			QRCode_stop_print();
		}
		#endif
	}
	return bit_size;
}

static int GenPolyF2_set_data(void * poly, int id, unsigned char bit)
{
	return GenPolyF2_set_val((GenPolyF2 *)poly, bit, id);
}

int QRCode_read_GenPolyF2(QRCode * code, GenPolyF2 * poly, int degree)
{
	if(! GenPolyF2_create(poly, PT_NORMAL, degree))
		return 0;
	return QRCode_read_data(code, poly, degree + 1, &GenPolyF2_set_data);
}

static int GenPolyF2_get_data(void * poly, int id, unsigned char * bit)
{
	if(id > GenPolyF2_real_degree((GenPolyF2 *)poly))
	{
		*bit = 0;
		return 1;
	}
	return GenPolyF2_get_val((GenPolyF2 *)poly, bit, id);
}

int QRCode_write_GenPolyF2(QRCode * code, const GenPolyF2 * poly, int degree)
{
	return QRCode_write_data(code, (void *)poly, degree + 1, &GenPolyF2_get_data);
}

static int CPoly_set_data(void * data, int current_id, unsigned char bit)
{
	CPoly * poly = NULL;
	int div;
	int cp_id;
	int val_id;
	
	poly = (CPoly *)data;
	if((div = GenPolyF2_real_degree(&(poly->based_poly))) <= 0)
		return 0;
	cp_id = current_id / div;
	val_id = div - (current_id % div) - 1;
	if(cp_id > poly->degree)
		return bit == 0;
	if(val_id > GenPolyF2_degree(CPoly_get_val(poly, cp_id)))
		return bit == 0;
	return GenPolyF2_set_val((GenPolyF2 *)CPoly_get_val(poly, cp_id), bit, val_id);
}

static int CPoly_get_data(void * data, int current_id, unsigned char * bit)
{
	CPoly * poly = NULL;
	int div;
	int cp_id;
	int val_id;
	
	poly = (CPoly *)data;
	if((div = GenPolyF2_real_degree(&(poly->based_poly))) <= 0)
		return 0;
	cp_id = current_id / div;
	val_id = div - (current_id % div) - 1;
	if(cp_id > poly->degree)
	{
		*bit = 0;
		return 1;
	}
	if(val_id > GenPolyF2_degree(CPoly_get_val(poly, cp_id)))
	{
		*bit = 0;
		return 1;
	}
	return GenPolyF2_get_val(CPoly_get_val(poly, cp_id), bit, val_id);
}

int QRCode_read_CPoly(QRCode * code, CPoly * poly, int degree)
{
	GenPolyF2 zero;
	int i;
	if(code->nb_codewords <= 0)
		return 0;
	if(! CPoly_create(poly, degree, &(code->rs.based_poly)))
		return 0;
	if(! GenPolyF2_create(&zero, PT_NORMAL, GenPolyF2_real_degree(&(poly->based_poly)) - 1))
	{
		CPoly_free(poly);
		return 0;
	}
	for(i = 0; i <= degree; ++i)
	{
		GenPolyF2_free((GenPolyF2 *)CPoly_get_val(poly, i));
		if(! CPoly_set_val(poly, &zero, i))
		{
			CPoly_free(poly);
			GenPolyF2_free(&zero);
			return 0;
		}
	}
	GenPolyF2_free(&zero);
	if(! QRCode_read_data(code, poly, (degree + 1) * GenPolyF2_real_degree(&(poly->based_poly)),
		&CPoly_set_data))
	{
		CPoly_free(poly);
		return 0;
	}
	CPoly_reverse_vals_order(poly);
	return 1;
}

int QRCode_write_CPoly(QRCode * code, CPoly * poly, int degree)
{
	CPoly_reverse_vals_order(poly);
	if(! QRCode_write_data(code, poly, (degree + 1) * GenPolyF2_real_degree(&(poly->based_poly)),
		&CPoly_get_data))
	{
		CPoly_reverse_vals_order(poly);
		return 0;
	}
	CPoly_reverse_vals_order(poly);
	return 1;
}

int QRCode_print(QRCode * code)
{
	SDL_Rect pos;
	if(screen != NULL)
		SDL_FreeSurface(screen);
	if((screen = SDL_SetVideoMode(code->surface->w, code->surface->h, 32, SDL_HWSURFACE)) == NULL)
	{
		fprintf(stderr, "Erreur SDL_SetVideoMode : %s\n", SDL_GetError());
		return 0;
	}
	pos.x = pos.y = 0;
	pos.w = code->surface->w;
	pos.h = code->surface->h;
	SDL_BlitSurface(code->surface, NULL, screen, &pos);
	SDL_Flip(screen);
	return 1;
}

int QRCode_print_zoom(QRCode * code, int scale)
{
	int x, y;
	int r, g, b;
	SDL_Rect rect;
	if(screen != NULL)
		SDL_FreeSurface(screen);
	if((screen = SDL_SetVideoMode(code->surface->w * scale, code->surface->h * scale, 32, SDL_HWSURFACE)) == NULL)
	{
		fprintf(stderr, "Erreur SDL_SetVideoMode : %s\n", SDL_GetError());
		return 0;
	}
	for(x = 0; x < code->surface->w; ++x)
	{
		for(y = 0; y < code->surface->h; ++y)
		{
			SDL_Surface_get_pixel(code->surface, x, y, &r, &g, &b, NULL);
			rect = get_rect(x * scale, y * scale, scale, scale);
			SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, r, g, b));
		}
	}
	SDL_Flip(screen);
	return 1;
}

void QRCode_print_autozoom(QRCode * code)
{
	QRCode_print_zoom(code, 20 - (code->version) / 2);
}

int QRCode_save(QRCode * code, const char * path)
{
	return SDL_SaveBMP(code->surface, path) == 0;
}

int QRCode_save_zoom(QRCode * code, int scale, const char * path)
{
	SDL_Surface * surface = NULL;
	int x, y;
	int r, g, b;
	SDL_Rect rect;
	
	if((surface = SDL_CreateRGBSurface(SDL_HWSURFACE, code->surface->w * scale, code->surface->h * scale, 32,
		rmask, gmask, bmask, amask)) == NULL)
		return 0;
	for(x = 0; x < code->surface->w; ++x)
	{
		for(y = 0; y < code->surface->h; ++y)
		{
			SDL_Surface_get_pixel(code->surface, x, y, &r, &g, &b, NULL);
			rect = get_rect(x * scale, y * scale, scale, scale);
			SDL_FillRect(surface, &rect, SDL_MapRGB(surface->format, r, g, b));
		}
	}
	if(SDL_SaveBMP(surface, path) != 0)
	{
		SDL_FreeSurface(surface);
		return 0;
	}
	SDL_FreeSurface(surface);
	return 1;
}

int QRCode_save_autozoom(QRCode * code, const char * path)
{
	return QRCode_save_zoom(code, 20 - (code->version) / 2, path);
}

int QRCode_wait_clic()
{
	SDL_Event event;
	int cont = 1;
	int success = 1;
	
	while(cont && success)
	{
		SDL_WaitEvent(&event);
		
		switch(event.type)
		{
			case SDL_QUIT:
			success = 0;
			break;
			
			case SDL_KEYDOWN:
			switch(event.key.keysym.sym)
			{
				case SDLK_ESCAPE:
				success = 0;
				break;
				
				default:
				break;
			}
			break;
			
			case SDL_MOUSEBUTTONUP:
			switch(event.button.button)
			{
				case SDL_BUTTON_LEFT:
				case SDL_BUTTON_RIGHT:
				cont = 0;
				break;
				
				default:
				break;
			}
			break;
			
			default:
			break;
		}
	}
	
	return success;
}

void QRCode_stop_print()
{
	if(screen == NULL)
		return;
	SDL_FreeSurface(screen);
	screen = NULL;
}

CorrectionLevel car_to_cl(char c)
{
	switch(c)
	{
		case 'L':
		return L;
		
		case 'M':
		return M;
		
		case 'Q':
		return Q;
		
		case 'H':
		return H;
		
		default:
		return -1;
	}
	return -1;
}

/* Graphics */

static void print_carre(SDL_Surface * surface, int x, int y, int size)
{
	SDL_Rect rect;
	rect = get_rect(x, y, 4 + size, 4 + size);
	SDL_FillRect(surface, &rect, SDL_MapRGB(surface->format, 0, 0, 0));
	rect = get_rect(x + 1, y + 1, 2 + size, 2 + size);
	SDL_FillRect(surface, &rect, SDL_MapRGB(surface->format, 255, 255, 255));
	rect = get_rect(x + 2, y + 2, size, size);
	SDL_FillRect(surface, &rect, SDL_MapRGB(surface->format, 0, 0, 0));
}

static void print_base(SDL_Surface * surface, int version)
{
	int i, j;
	SDL_Rect rect;
	
	SDL_LockSurface(surface);
	rect = get_rect(0, 0, 9, 9);
	SDL_FillRect(surface, &rect, SDL_MapRGB(surface->format, 255, 255, 255));
	rect = get_rect(surface->w - 8, 0, 8, 9);
	SDL_FillRect(surface, &rect, SDL_MapRGB(surface->format, 255, 255, 255));
	rect = get_rect(0, surface->h - 8, 9, 8);
	SDL_FillRect(surface, &rect, SDL_MapRGB(surface->format, 255, 255, 255));
	rect = get_rect(6, 0, 1, surface->h);
	SDL_FillRect(surface, &rect, SDL_MapRGB(surface->format, 255, 255, 255));
	rect = get_rect(0, 6, surface->w, 1);
	SDL_FillRect(surface, &rect, SDL_MapRGB(surface->format, 255, 255, 255));
	print_carre(surface, 0, 0, 3);
	print_carre(surface, surface->w - 7, 0, 3);
	print_carre(surface, 0, surface->h - 7, 3);
	SDL_UnlockSurface(surface);
	
	SDL_Surface_set_pixel(surface, 8, surface->h - 8, 0, 0, 0, 255);
	for(i = 8; i < surface->h - 8; i += 2)
	{
		SDL_Surface_set_pixel(surface, 6, i, 0, 0, 0, 255);
		SDL_Surface_set_pixel(surface, i, 6, 0, 0, 0, 255);
	}
	
	if(version <= 1)
		return;
	
	if(version >= 2 && version <= 6)
	{
		SDL_LockSurface(surface);
		print_carre(surface, 4 * (version + 2), 4 * (version + 2), 1);
		SDL_UnlockSurface(surface);
	}
	if(version >= 7 && version <= 13)
	{
		SDL_LockSurface(surface);
		for(i = 0; i < 3; ++i)
		{
			for(j = 0; j < 3; ++j)
			{
				if(i == 0 && (j == 0 || j == 2))
					continue;
				if(i == 2 && j == 0)
					continue;
				
				print_carre(surface, 4 + (i * 2) * (version + 1), 4 + (j * 2) * (version + 1), 1);
			}
		}
		SDL_UnlockSurface(surface);
	}
}

int get_nb_codewords(int version, CorrectionLevel cl)
{
	switch(version)
	{
		case 1: return 26;
		case 2: return 44;
		default: break;
	}
	return 0;
}

int get_nb_error_codewords(int version, CorrectionLevel cl)
{
	switch(version)
	{
		case 1:
		switch(cl)
		{
			case L:		return 7;
			case M:		return 10;
			case Q:		return 13;
			case H:		return 17;
			default:	break;
		}
		break;
		
		case 2:
		switch(cl)
		{
			case L:		return 10;
			case M:		return 16;
			case Q:		return 22;
			case H:		return 28;
			default:	break;
		}
		break;
		
		default:
		break;
	}
	return 0;
}

int get_capacity(int version, CorrectionLevel cl)
{
	return get_nb_codewords(version, cl) - get_nb_error_codewords(version, cl);
}

void QRCode_print_align(QRCode * code)
{
	print_base(code->surface, code->version);
}

int QRCode_create_version(QRCode * code, int version)
{
	if(version <= 0)
	{
		fprintf(stderr, "Version non prise en charge\n");
		return 0;
	}
	if(! QRCode_create(code, 17 + (4 * version)))
		return 0;
	code->version = version;
	return 1;
}

int QRCode_set_format(QRCode * code, CorrectionLevel cl, int mask_id)
{
	GenPolyF2 based_poly;
	code->cl = cl;
	code->mask_id = mask_id;
	if(! GenPolyF2_from_bin_str(&based_poly, "100011101", NULL))
		return 0;
	if(! RS_create(&(code->rs), &based_poly, get_nb_error_codewords(code->version, cl)))
	{
		GenPolyF2_free(&based_poly);
		return 0;
	}
	GenPolyF2_free(&based_poly);
	code->nb_codewords = get_nb_codewords(code->version, cl);
	return 1;
}

static int format_move(int id, SDL_Rect * coord, QRCode * code, void * user_data)
{
	if(id < 8)
	{
		coord->x = 8;
		coord->y = id;
		if(id >= 6)
			++(coord->y);
	}
	else if(id < 15)
	{
		coord->x = 8 - (id - 7);
		coord->y = 8;
		if(id >= 9)
			--(coord->x);
	}
	else if(id - 15 < 8)
	{
		coord->x = code->surface->w - 1 - (id - 15);
		coord->y = 8;
	}
	else if(id - 15 < 15)
	{
		coord->x = 8;
		coord->y = code->surface->h - 1 - 7 + (id - 15 - 7);
	}
	else
		return 0;
	return 1;
}

static int format_end(QRCode * code, void * user_data)
{
	if(code->current_id < *((int *)user_data))
		return 0;
	return 1;
}

static int format_init(QRCode * code, void * user_data)
{
	code->current_id = 0;
	return 1;
}

int QRCode_print_format(QRCode * code)
{
	GenPolyF2 format;
	GenPolyF2 message;
	GenPolyF2 gen;
	GenPolyF2 xor_mask;
	BCH bch;
	int i;
	int format_len = 30;
	int write_data = 0;
	
	#ifdef ANIM
	at_end ^= 1;
	#endif
	
	QRCode_print_align(code);
	if(! QRCode_set_move(code, NULL, &format_move, &format_end, &format_init, &format_len))
		return 0;
	if(! GenPolyF2_create(&format, PT_NORMAL, 5))
		return 0;
	for(i = 0; i < 3; ++i)
		GenPolyF2_set_val(&format, ((code->mask_id) & (1 << i)) != 0, i);
	for(i = 0; i < 2; ++i)
		GenPolyF2_set_val(&format, ((code->cl) & (1 << i)) != 0, i + 3);
	if(! GenPolyF2_from_bin_str(&gen, "10011", NULL))
	{
		GenPolyF2_free(&format);
		return 0;
	}
	if(! BCH_create(&bch, &gen, 3))
	{
		GenPolyF2_free(&format);
		GenPolyF2_free(&gen);
		return 0;
	}
	GenPolyF2_free(&gen);
	if(! BCH_code(&bch, &message, &format))
	{
		GenPolyF2_free(&format);
		BCH_free(&bch);
		return 0;
	}
	GenPolyF2_free(&format);
	BCH_free(&bch);
	if(! GenPolyF2_from_bin_str(&xor_mask, "101010000010010", NULL))
	{
		GenPolyF2_free(&message);
		return 0;
	}
	if(! GenPolyF2_self_add(&message, &xor_mask))
	{
		GenPolyF2_free(&xor_mask);
		GenPolyF2_free(&message);
		return 0;
	}
	GenPolyF2_free(&xor_mask);
	
	for(i = 0; i < 2; ++i)
		write_data += QRCode_write_GenPolyF2(code, &message, 14);
	/*GenPolyF2_print_bin(&message);*/
	GenPolyF2_free(&message);
	return format_len == write_data;
}

int QRCode_get_format(QRCode * code)
{
	GenPolyF2 format;
	GenPolyF2 get_message[2];
	GenPolyF2 gen;
	GenPolyF2 xor_mask;
	BCH bch;
	int i;
	int format_len = 30;
	int read_data = 0;
	int select_message = -1;
	unsigned char val;
	
	if(! QRCode_set_move(code, NULL, &format_move, &format_end, &format_init, &format_len))
		return 0;
		
	if(! GenPolyF2_from_bin_str(&gen, "10011", NULL))
		return 0;
	if(! BCH_create(&bch, &gen, 3))
	{
		GenPolyF2_free(&gen);
		return 0;
	}
	GenPolyF2_free(&gen);
	
	if(! GenPolyF2_from_bin_str(&xor_mask, "101010000010010", NULL))
	{
		BCH_free(&bch);
		return 0;
	}
	for(i = 0; i < 2; ++i)
	{
		read_data += QRCode_read_GenPolyF2(code, get_message + i, 14);
		if(! GenPolyF2_self_add(get_message + i, &xor_mask))
		{
			BCH_free(&bch);
			GenPolyF2_free(&xor_mask);
			if(i == 1)
				GenPolyF2_free(get_message);
			return 0;
		}
		/*GenPolyF2_print_bin(get_message + i);*/
	}
	GenPolyF2_free(&xor_mask);
	for(i = 0; i < 2; ++i)
	{
		if(! BCH_is_error(&bch, get_message + i))
			select_message = i;
	}
	if(select_message == -1)
		select_message = 0;
	if(! BCH_decode(&bch, &format, get_message + select_message))
	{
		BCH_free(&bch);
		for(i = 0; i < 2; ++i)
			GenPolyF2_free(get_message + i);
	}
	BCH_free(&bch);
	for(i = 0; i < 2; ++i)
		GenPolyF2_free(get_message + i);
	code->mask_id = 0;
	for(i = 0; i < 3; ++i)
	{
		if(! GenPolyF2_get_val(&format, &val, i))
			continue;
		code->mask_id |= ((val != 0) << i);
	}
	code->cl = 0;
	for(i = 0; i < 2; ++i)
	{
		if(! GenPolyF2_get_val(&format, &val, i + 3))
			continue;
		code->cl |= ((val != 0) << i);
	}
	GenPolyF2_free(&format);
	printf("correct %c, mask %d\n", *("MLHQ" + code->cl), code->mask_id);
	/*print_area(code);
	QRCode_wait_clic();
	QRCode_stop_print();*/
	if(! QRCode_set_format(code, code->cl, code->mask_id))
		return 0;
	return 1;
}

typedef enum
{
	DMM_UP,
	DMM_DOWN
} DataMoveMode;

typedef struct _DataMoveInfo DataMoveInfo;
struct _DataMoveInfo
{
	DataMoveMode direction;
};

static int is_message_end_qrdata_move(QRCode * code, void * user_data)
{
	if((code->current_id) / 8 >= code->nb_codewords)
		return 1;
	return 0;
}

static int can_move_on(int x, int y, QRCode * code)
{
	SDL_Surface * surface = NULL;
	surface = (code->area != NULL) ? (code->area) : (code->surface);
	return SDL_Surface_is_free_bit(surface, x, y);
}

static int first_free(int x, int y, int add, QRCode * code)
{
	if(x < 0 || x >= code->surface->w)
		return -1;
	while(y >= 0 && y < code->surface->h)
	{
		if(can_move_on(x, y, code))
			return y;
		y += add;
	}
	return -1;
}

static int get_coord_by_id_qrdata_move(int id, SDL_Rect * coord, QRCode * code, void * user_data)
{
	DataMoveInfo * dmm = NULL;
	int pline;
	int lval;
	int sub_lval;
	int cval;
	int signe;
	int cmp;
	if((dmm = (DataMoveInfo *)user_data) == NULL)
		return 0;
	if(id == 0)
	{
		coord->x = code->surface->w - 1;
		coord->y = code->surface->h - 1;
		dmm->direction = DMM_UP;
		return 1;
	}
	
	if(coord->x < 6)
		pline = coord->x % 2;
	else
		pline = (coord->x + 1) % 2;
	signe = (dmm->direction == DMM_UP) ? (-1) : (1);
	if(pline == 1)
		lval = first_free(cval = (coord->x - (1 + (coord->x - 1 == 6))), coord->y, signe, code);
	else
		lval = first_free(cval = (coord->x + 1), coord->y + signe, signe, code);
	if((sub_lval = first_free(coord->x, coord->y + signe, signe, code)) >= 0)
	{
		if(signe > 0)
			cmp = sub_lval < lval;
		else
			cmp = sub_lval > lval;
		if(cmp)
		{
			cval = coord->x;
			lval = sub_lval;
		}
	}
	/*printf("%d => %d | (%d, %d)\n", id, pline, cval, lval);*/
	if(lval >= 0)
	{
		coord->x = cval;
		coord->y = lval;
		return 1;
	}
	cval = coord->x - pline - (1 + (coord->x - 1 == 6));
	if((lval = first_free(cval, (dmm->direction == DMM_UP) ? (0) : (code->surface->h - 1), -signe, code)) < 0)
		return 0;
	dmm->direction = (dmm->direction == DMM_UP) ? (DMM_DOWN) : (DMM_UP);
	/*printf("(%d, %d)\n", cval, lval);*/
	coord->x = cval;
	coord->y = lval;
	
	return 1;
}

static int init_data_qrdata_move(QRCode * code, void * user_data)
{
	DataMoveInfo * dmm = NULL;
	code->current_id = 0;
	if((dmm = (DataMoveInfo *)user_data) == NULL)
		return 0;
	dmm->direction = DMM_UP;
	code->current_coord.x = code->surface->w - 1;
	code->current_coord.y = code->surface->h - 1;
	return 1;
}

static void QRCode_code_fill_blank(QRCode * code)
{
	int x, y;
	int r, g, b;
	for(x = 0; x < code->surface->w; ++x)
	{
		for(y = 0; y < code->surface->h; ++y)
		{
			SDL_Surface_get_pixel(code->surface, x, y, &r, &g, &b, NULL);
			if(r == 127 && g == 127 && b == 127)
				SDL_Surface_set_pixel(code->surface, x, y, 255, 255, 255, 255);
		}
	}
}

int QRCode_code_data(QRCode * code, const unsigned char * data)
{
	CPoly data_poly;
	DataMoveInfo dmm;
	
	if(! RS_code_chars(&(code->rs), code->nb_codewords, data, &data_poly))
		return 0;
	/*CPoly_print_hexa(&data_poly);*/
	if(! QRCode_set_move(code, NULL, &get_coord_by_id_qrdata_move, &is_message_end_qrdata_move, &init_data_qrdata_move, &dmm))
	{
		CPoly_free(&data_poly);
		return 0;
	}
	if(! QRCode_write_CPoly(code, &data_poly, code->nb_codewords))
	{
		CPoly_free(&data_poly);
		return 0;
	}
	QRCode_code_fill_blank(code);
	CPoly_free(&data_poly);
	return 1;
}

int QRCode_decode_data(QRCode * code, unsigned char ** data)
{
	CPoly data_poly;
	DataMoveInfo dmm;
	CPoly extract_data;
	
	SDL_FillRect(code->area, NULL, SDL_MapRGB(code->area->format, 127, 127, 127));
	print_base(code->area, code->version);
	QRCode_apply_mask(code);
	if(! QRCode_set_move(code, NULL, &get_coord_by_id_qrdata_move, &is_message_end_qrdata_move, &init_data_qrdata_move, &dmm))
		return 0;
	/*SDL_SaveBMP(code->surface, "media/no_mask.bmp");*/
	if(! QRCode_read_CPoly(code, &data_poly, code->nb_codewords))
		return 0;
	/*CPoly_print_hexa(&data_poly);*/
	code->rs.nb_words = code->nb_codewords;
	if(! RS_decode(&(code->rs), &data_poly, &extract_data))
	{
		CPoly_free(&data_poly);
		return 0;
	}
	CPoly_free(&data_poly);
	
	if(! CPoly_to_data(data, NULL, &extract_data))
	{
		CPoly_free(&extract_data);
		return 0;
	}
	CPoly_free(&extract_data);
	QRCode_apply_mask(code);
	
	return 1;
}

static int mask_val(int id, int j, int i)
{
	switch(id)
	{
		case 0:
		return (i + j) % 2 == 0;
		
		case 1:
		return i % 2 == 0;
		
		case 2:
		return j % 3 == 0;
		
		case 3:
		return (i + j) % 3 == 0;
		
		case 4:
		return ((i / 2) + (j / 3)) % 2 == 0;
		
		case 5:
		return (((i * j) % 2) + ((i * j) % 3)) == 0;
		
		case 6:
		return (((i * j) % 2) + ((i * j) % 3)) % 2 == 0;
		
		case 7:
		return (((i + j) % 2) + ((i * j) % 3)) % 2 == 0;
		
		default:
		break;
	}
	return 0;
}

void QRCode_apply_mask(QRCode * code)
{
	int i, j;
	int val;
	int r, g, b;
	for(i = 0; i < code->surface->w; ++i)
	{
		for(j = 0; j < code->surface->h; ++j)
		{
			if((val = mask_val(code->mask_id, i, j)) == 0)
				continue;
			val *= 255;
			SDL_Surface_get_pixel(code->surface, i, j, &r, &g, &b, NULL);
			SDL_Surface_set_pixel(code->surface, i, j, r ^ val, g ^ val, b ^ val, 255);
		}
	}
}

static int get_color_axe_x(const SDL_Surface * image, int l, int i, int * r, int * g, int * b)
{
	if(i < 0 || i >= image->h)
		return 0;
	SDL_Surface_get_pixel(image, l, i, r, g, b, NULL);
	return 1;
}

static int get_color_axe_y(const SDL_Surface * image, int l, int i, int * r, int * g, int * b)
{
	if(i < 0 || i >= image->w)
		return 0;
	SDL_Surface_get_pixel(image, i, l, r, g, b, NULL);
	return 1;
}

static int color_strict_equal(int r1, int g1, int b1, int r2, int g2, int b2)
{
	return r1 == r2 && g1 == g2 && b1 == b2;
}

static int is_color_line(const SDL_Surface * image, int line, int r, int g, int b, int (*get_color)(const SDL_Surface *, int l, int i, int * r, int * g, int * b), int (*equal)(int r1, int g1, int b1, int r2, int g2, int b2))
{
	int i;
	int gr = 0, gg = 0, gb = 0;
	for(i = 0; (*get_color)(image, line, i, &gr, &gg, &gb); ++i)
	{
		if(! (*equal)(gr, gg, gb, r, g, b))
			return 0;
	}
	return 1;
}

static int is_color(const SDL_Surface * image, int x, int y, int r, int g, int b)
{
	int gr, gg, gb;
	if(x < 0 || y < 0 || x >= image->w || y >= image->h)
		return 0;
	SDL_Surface_get_pixel(image, x, y, &gr, &gg, &gb, NULL);
	return color_strict_equal(r, g, b, gr, gg, gb);
}

static int get_size(const SDL_Surface * image, SDL_Rect area)
{
	int square_size = 0;
	int pos;
	
	for(pos = 0; is_color(image, area.x + pos, area.y + pos, 0, 0, 0); ++pos);
	for( ; is_color(image, area.x + pos, area.y + pos, 255, 255, 255); ++pos);
	for(square_size = 0; is_color(image, area.x + pos, area.y + pos, 0, 0, 0); ++pos, ++square_size);
	
	/*printf("square size %d\n", square_size);*/
	
	if(square_size <= 0)
		return 0;
	return (area.w * 3) / square_size;
}

int Image_to_matrix(const SDL_Surface * image, SDL_Surface ** matrix)
{
	SDL_Surface * mat = NULL;
	SDL_Rect area;
	int x_min, x_max, y_min, y_max;
	int size;
	int x, y;
	int r, g, b;
	int val;
	
	for(x_min = 0; x_min < image->w && is_color_line(image, x_min, 255, 255, 255, &get_color_axe_x, &color_strict_equal); ++x_min);
	for(y_min = 0; y_min < image->h && is_color_line(image, y_min, 255, 255, 255, &get_color_axe_y, &color_strict_equal); ++y_min);
	for(x_max = image->w; x_max > 0 && is_color_line(image, x_max - 1, 255, 255, 255, &get_color_axe_x, &color_strict_equal); --x_max);
	for(y_max = image->h; y_max > 0 && is_color_line(image, y_max - 1, 255, 255, 255, &get_color_axe_y, &color_strict_equal); --y_max);
	/*printf("%d %d %d %d\n", x_min, y_min, x_max, y_max);*/
	
	area = get_rect(x_min, y_min, x_max - x_min, y_max - y_min);
	if((size = get_size(image, area)) <= 0)
		return 0;
	/*printf("size %d\n", size);*/
	if((mat = SDL_CreateRGBSurface(SDL_HWSURFACE, size, size, 32, rmask, gmask, bmask, amask)) == NULL)
		return 0;
	for(x = 0; x < size; ++x)
	{
		for(y = 0; y < size; ++y)
		{
			SDL_Surface_get_pixel(image, area.x + (x * area.w) / size, area.y + (y * area.h) / size, &r, &g, &b, NULL);
			val = (r + g + b > 127 * 3) ? 255 : 0;
			SDL_Surface_set_pixel(mat, x, y, val, val, val, 255);
		}
	}
	*matrix = mat;
	return 1;
}