#include "reed-solomon.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "peterson_solve.h"

static int CPoly_get_prod(CPoly * res, const GenPolyF2 * based_poly, int power)
{
	GenPolyF2 tmp[2];
	CPoly next;
	if(power < 0)
		return 0;
	if(! GenPolyF2_create_single(tmp + 1, PT_NORMAL, 0)
	|| ! GenPolyF2_create_single(tmp + 0, PT_NORMAL, 1))
		return 0;
	if(! GenPolyF2_self_power(tmp + 0, power, based_poly))
		goto error_zone;
	if(! CPoly_create_by_tab(res, tmp, 2, based_poly))
		goto error_zone;
	GenPolyF2_free(tmp + 0);
	GenPolyF2_free(tmp + 1);
	if(power == 0)
		return 1;
	if(! CPoly_get_prod(&next, based_poly, power - 1))
	{
		CPoly_free(res);
		return 0;
	}
	if(! CPoly_self_mul(res, &next))
	{
		CPoly_free(res);
		return 0;
	}
	
	return 1;
	
	error_zone:;
	
	GenPolyF2_free(tmp + 0);
	GenPolyF2_free(tmp + 1);
	
	return 0;
}

int RS_create(RS * rs, const GenPolyF2 * based_poly, int nb_correct)
{
	if(! CPoly_get_prod(&(rs->g), based_poly, nb_correct - 1))
		return 0;
	/*CPoly_print(&(rs->g));*/
	if(! GenPolyF2_copy(&(rs->based_poly), based_poly))
	{
		CPoly_free(&(rs->g));
		return 0;
	}
	rs->nb_correct = nb_correct;
	rs->nb_words = 0;
	return 1;
}

void RS_free(RS * rs)
{
	CPoly_free(&(rs->g));
	GenPolyF2_free(&(rs->based_poly));
	rs->nb_correct = 0;
}

int RS_code(RS * rs, const CPoly * data_poly, CPoly * coded_poly)
{
	GenPolyF2 tmp;
	CPoly reminder;
	if(! GenPolyF2_create_single(&tmp, PT_NORMAL, 0))
		return 0;
	if(! CPoly_create(coded_poly, rs->nb_correct, &(rs->based_poly)))
	{
		GenPolyF2_free(&tmp);
		return 0;
	}
	if(! CPoly_set_val(coded_poly, &tmp, rs->nb_correct))
	{
		GenPolyF2_free(&tmp);
		CPoly_free(coded_poly);
		return 0;
	}
	GenPolyF2_free(&tmp);
	if(! CPoly_self_mul(coded_poly, data_poly))
	{
		CPoly_free(coded_poly);
		return 0;
	}
	if(! CPoly_euclid_div(NULL, &reminder, coded_poly, &(rs->g)))
	{
		CPoly_free(coded_poly);
		return 0;
	}
	if(! CPoly_self_add(coded_poly, &reminder))
	{
		CPoly_free(coded_poly);
		CPoly_free(&reminder);
		return 0;
	}
	CPoly_free(&reminder);
	return 1;
}

static int RS_get_syndroms(RS * rs, const CPoly * coded_poly, GenPolyF2 ** syndroms)
{
	GenPolyF2 * tab = NULL;
	GenPolyF2 tmp;
	int i;
	
	if((tab = (GenPolyF2 *)calloc(rs->nb_correct, sizeof(GenPolyF2))) == NULL)
		return 0;
	for(i = 0; i < rs->nb_correct; ++i)
	{
		if(! GenPolyF2_create_single(&tmp, PT_NORMAL, i))
			goto error_zone;
		if(! CPoly_eval(tab + i, coded_poly, &tmp))
		{
			GenPolyF2_free(&tmp);
			goto error_zone;
		}
	}
	
	*syndroms = tab;
	
	return 1;
	
	error_zone:;
	
	for(--i; i >= 0; --i)
		GenPolyF2_free(tab + i);
	free(tab);
	return 0;
}

static int GenPolyF2_all_zero(GenPolyF2 * tab, int len)
{
	int i;
	for(i = 0; i < len; ++i)
	{
		if(! GenPolyF2_equals_vals(tab + i, 0, -1))
			return 0;
	}
	return 1;
}

static int make_omega_and_deriv(RS * rs, CPoly * locator, GenPolyF2 * syndroms, CPoly * omega, CPoly * deriv)
{
	CPoly syndrom_poly;
	CPoly x2t;
	GenPolyF2 tmp;
	if(! GenPolyF2_create_single(&tmp, PT_NORMAL, 0))
		return 0;
	if(! CPoly_create(&x2t, rs->nb_correct - 1, &(rs->based_poly)))
	{
		GenPolyF2_free(&tmp);
		return 0;
	}
	if(! CPoly_set_val(&x2t, &tmp, rs->nb_correct - 1))
	{
		GenPolyF2_free(&tmp);
		CPoly_free(&x2t);
		return 0;
	}
	GenPolyF2_free(&tmp);
	if(! CPoly_create_by_tab(&syndrom_poly, syndroms, rs->nb_correct - 1, &(rs->based_poly)))
	{
		CPoly_free(&x2t);
		return 0;
	}
	/*printf("syndroms: ");
	CPoly_print_hexa(&syndrom_poly);*/
	/*CPoly_print_hexa(&x2t);*/
	if(! CPoly_self_mul(&syndrom_poly, locator))
	{
		CPoly_free(&syndrom_poly);
		CPoly_free(&x2t);
		return 0;
	}
	if(! CPoly_euclid_div(NULL, omega, &syndrom_poly, &x2t))
	{
		CPoly_free(&syndrom_poly);
		CPoly_free(&x2t);
		return 0;
	}
	CPoly_free(&syndrom_poly);
	CPoly_free(&x2t);
	if(! CPoly_formal_derivative(deriv, locator))
	{
		CPoly_free(omega);
		return 0;
	}
	/*printf("omega :");
	CPoly_print_hexa(omega);
	printf("deriv: ");
	CPoly_print_hexa(deriv);*/
	return 1;
}

static int error_val(RS * rs, CPoly * omega, CPoly * deriv, GenPolyF2 * racine, GenPolyF2 * final_mul, GenPolyF2 * val)
{
	GenPolyF2 sup;
	GenPolyF2 sub;
	GenPolyF2 inv_sub;
	
	if(! CPoly_eval(&sup, omega, racine))
		return 0;
	if(! CPoly_eval(&sub, deriv, racine))
	{
		GenPolyF2_free(&sup);
		return 0;
	}
	if(! GenPolyF2_invert(&inv_sub, &sub, &(rs->based_poly)))
	{
		GenPolyF2_free(&sup);
		GenPolyF2_free(&sub);
		return 0;
	}
	GenPolyF2_free(&sub);
	if(! GenPolyF2_mul(val, &sup, &inv_sub, &(rs->based_poly)))
	{
		GenPolyF2_free(&sup);
		GenPolyF2_free(&inv_sub);
		return 0;
	}
	GenPolyF2_free(&sup);
	GenPolyF2_free(&inv_sub);
	if(! GenPolyF2_self_mul(val, final_mul, &(rs->based_poly)))
	{
		GenPolyF2_free(val);
		return 0;
	}
	return 1;
}

static int RS_correct_errors(RS * rs, CPoly * locator, CPoly * coded_poly, GenPolyF2 * syndroms)
{
	GenPolyF2 tmp;
	GenPolyF2 inv;
	GenPolyF2 racine;
	CPoly omega;
	CPoly deriv;
	GenPolyF2 val;
	int i;
	if(! make_omega_and_deriv(rs, locator, syndroms, &omega, &deriv))
		return 0;
	for(i = 0; i < rs->nb_words; ++i)
	{
		if(! GenPolyF2_create_single(&racine, PT_NORMAL, i))
			goto error_zone;
		if(! GenPolyF2_invert(&inv, &racine, &(rs->based_poly)))
		{
			GenPolyF2_free(&racine);
			goto error_zone;
		}
		
		if(! CPoly_eval(&tmp, locator, &inv))
		{
			GenPolyF2_free(&racine);
			GenPolyF2_free(&inv);
			goto error_zone;
		}
		if(GenPolyF2_equals_vals(&tmp, 0, -1))
		{
			printf("Erreur en %d\n", i);
			if(! error_val(rs, &omega, &deriv, &inv, &racine, &val))
			{
				GenPolyF2_free(&racine);
				GenPolyF2_free(&inv);
				GenPolyF2_free(&tmp);
				goto error_zone;
			}
			CPoly_add_val(coded_poly, &val, i);
			/*GenPolyF2_print_hexa(&val);*/
			/*printf("\n");*/
			GenPolyF2_free(&val);
		}
		GenPolyF2_free(&racine);
		GenPolyF2_free(&inv);
		GenPolyF2_free(&tmp);
	}
	return 1;
	
	error_zone:;
	
	CPoly_free(&omega);
	CPoly_free(&deriv);
	return 0;
}

int RS_decode(RS * rs, CPoly * coded_poly, CPoly * data_poly)
{
	GenPolyF2 * syndroms = NULL;
	CPoly locator;
	int i;
	
	if(! RS_get_syndroms(rs, coded_poly, &syndroms))
		return 0;
	/*print_gen_tab(syndroms, rs->nb_correct);*/
	if(GenPolyF2_all_zero(syndroms, rs->nb_correct))
	{
		free_gen_tab(&syndroms, rs->nb_correct);
		printf("Pas d'erreurs\n");
		goto read_message;
	}
	printf("Erreurs détectées\n");
	if(! CPoly_locator(&locator, &(rs->based_poly), syndroms, rs->nb_correct / 2))
	{
		free_gen_tab(&syndroms, rs->nb_correct);
		fprintf(stderr, "Erreur de récupération du polynôme localisateur d'erreurs\n");
		return 0;
	}
	/*printf("locator: ");
	CPoly_print_hexa(&locator);*/
	if(! RS_correct_errors(rs, &locator, coded_poly, syndroms))
	{
		free_gen_tab(&syndroms, rs->nb_correct);
		CPoly_free(&locator);
		fprintf(stderr, "Erreur correction des erreurs\n");
		return 0;
	}
	CPoly_free(&locator);
	free_gen_tab(&syndroms, rs->nb_correct);
	
	read_message:;
	
	/*CPoly_print_hexa(coded_poly);*/
	
	if(! CPoly_create(data_poly, rs->nb_words - rs->nb_correct - 1, &(rs->based_poly)))
		return 0;
	for(i = 0; i < rs->nb_words - rs->nb_correct; ++i)
		CPoly_set_val(data_poly, CPoly_get_val(coded_poly, i + rs->nb_correct), i);
		
	/*CPoly_print_hexa(data_poly);*/
	
	return 1;
}

void RS_print(const RS * rs)
{
	printf("Reed-Solomon %d codewords correct:\n", rs->nb_correct);
	printf("reference : ");
	GenPolyF2_print_bin(&(rs->based_poly));
	printf("generator : ");
	CPoly_print_hexa(&(rs->g));
}

static int GenPolyF2_from_car(GenPolyF2 * res, unsigned char car)
{
	int i;
	if(! GenPolyF2_create(res, PT_NORMAL, 7))
		return 0;
	for(i = 0; i < 8; ++i)
	{
		if(! GenPolyF2_set_val(res, (car & (1 << i)) != 0, i))
		{
			GenPolyF2_free(res);
			return 0;
		}
	}
	return 1;
}

int RS_code_chars(RS * rs, int nb_codewords, const unsigned char * chaine, CPoly * coded_poly)
{
	GenPolyF2 * tab = NULL;
	int i;
	int size;
	int success;
	int len;
	CPoly msg_poly;
	
	if(nb_codewords <= rs->nb_correct)
		return 0;
	
	size = nb_codewords - rs->nb_correct;
	len = strlen((char *)chaine);
	
	if((tab = (GenPolyF2 *)calloc(size, sizeof(GenPolyF2))) == NULL)
		return 0;
	for(i = 0; i < size; ++i)
	{
		if(! GenPolyF2_from_car(tab + i, (size - i - 1 < len) ? (*(chaine + size - i - 1)) : 0))
			goto error_zone;
	}
	if(! CPoly_create_by_tab(&msg_poly, tab, size, &(rs->based_poly)))
		goto error_zone;
	success = RS_code(rs, &msg_poly, coded_poly);
	
	CPoly_free(&msg_poly);
	for(i = 0; i < size; ++i)
		GenPolyF2_free(tab + i);
	free(tab);
	
	return success;
	
	error_zone:;
	
	for(--i; i >= 0; --i)
		GenPolyF2_free(tab + i);
	free(tab);
	
	return 0;
}

int hexa_str_to_chars(const char * hexa_str, char ** chaine)
{
	int len;
	char * res = NULL;
	int i;
	unsigned char val;
	
	len = strlen(hexa_str);
	if((res = (char *)calloc((len / 2) + 1, sizeof(char))) == NULL)
		return 0;
	for(i = 0; i < len; ++i)
	{
		if(*(hexa_str + i) >= '0' && *(hexa_str + i) <= '9')
			val = (unsigned char)(*(hexa_str + i)) - '0';
		else if(*(hexa_str + i) >= 'a' && *(hexa_str + i) <= 'f')
			val = (unsigned char)(*(hexa_str + i)) - 'a' + 10;
		else if(*(hexa_str + i) >= 'A' && *(hexa_str + i) <= 'F')
			val = (unsigned char)(*(hexa_str + i)) - 'A' + 10;
		else
		{
			free(res);
			return 0;
		}
		*((unsigned char *)(res + (i / 2))) |= (val << ((i % 2 == 0) * 4));
	}
	*chaine = res;
	return 1;
}