#include "test.h"

#include "color_print.h"
#include "poly.h"
#include "poly_int.h"
#include "generic_poly.h"
#include "bch.h"
#include "find_primes.h"
#include "test_control.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define RGB_GOOD 53, 134, 63
#define RGB_BAD 227, 41, 41
#define RGB_TITLE 69, 106, 141

#define CONTINUE 1
#define STOP_AFTER_ERROR 0

#define TEST_TRIALS 4242
#define TEST_SIZE_INT 32
#define TEST_SIZE_NORMAL 65536
#define TEST_SIZE_MAX TEST_SIZE_NORMAL
#define TEST_SIZE(type) (((type) == PT_NORMAL) ? (TEST_SIZE_NORMAL) : (TEST_SIZE_INT))
#define PT_random() ((rand() % 2) ? (PT_NORMAL) : (PT_INT))

static int make_test(const char * name, Test list[], int list_len, TextModif modif, int continue_after_error)
{
	int i;
	int success = 1;
	
	rgbprintf(modif, RGB_TITLE, "%s:\n", name);
	
	for(i = 0; i < list_len; ++i)
	{
		if(! (*(list[i].function))())
		{
			rgbprintf(modif, RGB_BAD, "%s : ERROR\n", list[i].name);
			success = 0;
			if(! continue_after_error)
				break;
		}
		else
			rgbprintf(modif, RGB_GOOD, "%s : OK\n", list[i].name);
	}
	
	return success;
}

int full_test()
{
	Test list[] = {	{"Full Test Poly", &full_test_poly}};
	const int list_len = 1;
	
	srand(time(NULL));
	
	return make_test("Full Test", list, list_len, TM_BOLD | TM_UNDERLINE, CONTINUE);
}

/* GenPolyF2 */

int full_test_poly()
{
	Test list[] = {	{"Poly Init", &test_poly_init},
			{"Poly Operations", &test_poly_operations}};
	const int list_len = 2;
	
	return make_test("Full Test Poly", list, list_len, TM_BOLD, STOP_AFTER_ERROR);
}

static int test_poly_init_create()
{
	int i;
	int deg;
	int success;
	GenPolyF2 poly;
	GenPolyF2 tab_poly[TEST_TRIALS];
	PolyType type;
	
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		type = PT_random();
		if(! GenPolyF2_create(&poly, type, deg = (1 + rand() % (TEST_SIZE(type) - 1))))
		{
			fprintf(stderr, "No Poly creation\n");
			return 0;
		}
		if(GenPolyF2_degree(&poly) != deg)
		{
			fprintf(stderr, "Incorrect degree\n");
			GenPolyF2_free(&poly);
			return 0;
		}
		GenPolyF2_free(&poly);
	}
	
	success = 1;
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		type = PT_random();
		if(! GenPolyF2_create(tab_poly + i, type, deg = (1 + rand() % (TEST_SIZE(type) - 1))))
		{
			fprintf(stderr, "No Poly creation in tab\n");
			success = 0;
			break;
		}
		if(GenPolyF2_degree(tab_poly + i) != deg)
		{
			fprintf(stderr, "Incorrect degree in tab\n");
			++i;
			success = 0;
			break;
		}
	}
	for(--i; i >= 0; --i)
		GenPolyF2_free(tab_poly + i);
	
	return success;
}

static int test_poly_init_modif_vals()
{
	int i;
	int j;
	int deg;
	unsigned char val;
	GenPolyF2 poly;
	unsigned char vals[TEST_SIZE_MAX];
	PolyType type;
	
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		type = PT_random();
		if(! GenPolyF2_create(&poly, type, deg = (1 + rand() % (TEST_SIZE(type) - 1))))
		{
			fprintf(stderr, "Init error\n");
			return 0;
		}
		for(j = 0; j <= deg; ++j)
		{
			vals[j] = rand() % 2;
			if(! GenPolyF2_set_val(&poly, vals[j], j))
			{
				fprintf(stderr, "Cannot set value\n");
				GenPolyF2_free(&poly);
				return 0;
			}
		}
		for(j = 0; j <= deg; ++j)
		{
			if(! GenPolyF2_get_val(&poly, &val, j))
			{
				fprintf(stderr, "Cannot get value\n");
				GenPolyF2_free(&poly);
				return 0;
			}
			if(val != vals[j])
			{
				fprintf(stderr, "Incorrect values %d %d [%d]\n", val, vals[j], j);
				GenPolyF2_free(&poly);
				return 0;
			}
		}
		GenPolyF2_free(&poly);
	}
	return 1;
}

static int make_alea_poly(GenPolyF2 * poly, unsigned char * vals, PolyType type, int deg)
{
	int i;
	unsigned char val;
	
	if(! GenPolyF2_create(poly, type, deg))
		return 0;
	for(i = 0; i <= deg; ++i)
	{
		val = rand() % 2;
		if(vals != NULL)
			vals[i] = val;
		if(! GenPolyF2_set_val(poly, val, i))
		{
			fprintf(stderr, "Cannot create an alea poly\n");
			GenPolyF2_free(poly);
			return 0;
		}
	}
	return 1;
}

static int is_tab_equal(const GenPolyF2 * poly, unsigned char * vals, int deg)
{
	int i;
	unsigned char val;
	for(i = 0; i <= deg; ++i)
	{
		if(! GenPolyF2_get_val(poly, &val, i) && vals[i] != 0)
		{
			fprintf(stderr, "Get error\n");
			return 0;
		}
		if(val != vals[i])
		{
			fprintf(stderr, "Differents values %d | %d [%d]\n", val, vals[i], i);
			return 0;
		}
	}
	return 1;
}

static int test_poly_equals()
{
	GenPolyF2 first;
	GenPolyF2 second;
	int i;
	int j;
	int deg;
	int is_eq;
	unsigned char vals_first[TEST_SIZE_MAX];
	unsigned char vals_second[TEST_SIZE_MAX];
	PolyType type;
	
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		type = PT_random();
		if(! GenPolyF2_create(&first, type, deg = (1 + rand() % (TEST_SIZE(type) - 1))))
		{
			fprintf(stderr, "Init error\n");
			return 0;
		}
		if(! GenPolyF2_create(&second, PT_NORMAL, deg + rand() % 42))
		{
			fprintf(stderr, "Init error\n");
			GenPolyF2_free(&first);
			return 0;
		}
		is_eq = 1;
		for(j = 0; j <= deg; ++j)
		{
			vals_first[j] = rand() % 2;
			if(rand() % (TEST_SIZE(type) / 2) == 0)
				vals_second[j] = rand() % 2;
			else
				vals_second[j] = vals_first[j];
			if(! GenPolyF2_set_val(&first, vals_first[j], j)
			|| ! GenPolyF2_set_val(&second, vals_second[j], j))
			{
				fprintf(stderr, "Cannot define values\n");
				goto error_zone;
			}
			if(vals_first[j] != vals_second[j])
				is_eq = 0;
		}
		if(GenPolyF2_equals(&first, &second) != is_eq)
		{
			fprintf(stderr, "Incorrect equality\n");
			goto error_zone;
		}
		GenPolyF2_free(&first);
		GenPolyF2_free(&second);
	}
	
	return 1;
	
	error_zone:;
	
	GenPolyF2_free(&first);
	GenPolyF2_free(&second);
	return 0;
}

static int test_poly_init_other_set()
{
	GenPolyF2 reference;
	GenPolyF2 poly;
	int i;
	int deg;
	unsigned char vals[TEST_SIZE_MAX];
	PolyType type;
	
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		if(! make_alea_poly(&reference, vals, PT_random(), deg = 10))
			return 0;
		if(! GenPolyF2_create(&poly, PT_random(), deg))
		{
			fprintf(stderr, "Init error\n");
			GenPolyF2_free(&reference);
			return 0;
		}
		if(! GenPolyF2_set_vals(&poly, vals[0], vals[1], vals[2], vals[3],
					vals[4], vals[5], vals[6], vals[7],
					vals[8], vals[9], vals[10], -1))
		{
			fprintf(stderr, "Error in multi set values\n");
			goto error_zone;
		}
		if(! GenPolyF2_equals(&reference, &poly))
		{
			fprintf(stderr, "Incorrect equality multi set\n");
			goto error_zone;
		}
		GenPolyF2_free(&poly);
		if(! GenPolyF2_create_by_vals(&poly, PT_random(), vals[0], vals[1], vals[2], vals[3],
					vals[4], vals[5], vals[6], vals[7],
					vals[8], vals[9], vals[10], -1))
		{
			fprintf(stderr, "Error in multi create values\n");
			GenPolyF2_free(&reference);
			return 0;
		}
		if(! GenPolyF2_equals(&reference, &poly))
		{
			fprintf(stderr, "Incorrect equality multi create\n");
			goto error_zone;
		}
		GenPolyF2_free(&poly);
		GenPolyF2_free(&reference);
	}
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		type = PT_random();
		if(! GenPolyF2_create(&reference, type, deg = (1 + rand() % (TEST_SIZE(PT_INT) - 1))))
		{
			fprintf(stderr, "Init error\n");
			return 0;
		}
		if(! GenPolyF2_set_val(&reference, 1, deg)
		|| ! GenPolyF2_create_single(&poly, PT_random(), deg))
		{
			fprintf(stderr, "No single coeff poly created\n");
			GenPolyF2_free(&reference);
			return 0;
		}
		if(! GenPolyF2_equals(&reference, &poly))
		{
			fprintf(stderr, "Incorrect equality\n");
			goto error_zone;
		}
		GenPolyF2_free(&poly);
		GenPolyF2_free(&reference);
	}
	
	return 1;
	
	error_zone:;
	
	GenPolyF2_free(&reference);
	GenPolyF2_free(&poly);
	return 0;
}

static int test_poly_init_copy()
{
	GenPolyF2 poly;
	GenPolyF2 copy;
	int i;
	int deg;
	PolyType type;
	
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		type = PT_random();
		if(! make_alea_poly(&poly, NULL, type, deg = (1 + rand() % (TEST_SIZE(type) - 1))))
			return 0;
		if(! GenPolyF2_copy(&copy, &poly))
		{
			fprintf(stderr, "Init error\n");
			GenPolyF2_free(&poly);
			return 0;
		}
		if(! GenPolyF2_equals(&copy, &poly))
		{
			fprintf(stderr, "Incorrect equality\n");
			GenPolyF2_free(&poly);
			GenPolyF2_free(&copy);
			return 0;
		}
		GenPolyF2_free(&poly);
		GenPolyF2_free(&copy);
	}
	
	return 1;
}

int test_poly_init()
{
	Test list[] = {	{"Poly create", &test_poly_init_create},
			{"Poly set/get", &test_poly_init_modif_vals},
			{"Poly equality", &test_poly_equals},
			{"Poly different sets", &test_poly_init_other_set},
			{"Poly copy", &test_poly_init_copy}};
	const int list_len = 5;
	
	return make_test("Poly Init", list, list_len, TM_NORMAL, STOP_AFTER_ERROR);
}

static int test_real_degree()
{
	GenPolyF2 poly;
	int i;
	int j;
	int deg;
	int real_deg;
	int max_deg;
	unsigned char vals[TEST_SIZE_MAX];
	PolyType type;
	
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		type = PT_random();
		if(! make_alea_poly(&poly, vals, type, deg = (1 + rand() % (TEST_SIZE(type) - 1))))
			return 0;
		if((real_deg = GenPolyF2_real_degree(&poly)) < 0)
		{
			fprintf(stderr, "Inval real degree\n");
			goto error_zone;
		}
		max_deg = 0;
		for(j = 0; j <= deg; ++j)
		{
			if(vals[j] == 1)
				max_deg = j;
		}
		if(max_deg != real_deg)
		{
			fprintf(stderr, "Differents degree found %d | %d\n", real_deg, max_deg);
			goto error_zone;
		}
		GenPolyF2_free(&poly);
	}
	
	return 1;
	
	error_zone:;
	
	GenPolyF2_free(&poly);
	return 0;
}

static int test_poly_op_add_aux(int (*function)(GenPolyF2 *, const GenPolyF2 *, const GenPolyF2 *))
{
	GenPolyF2 first;
	GenPolyF2 second;
	GenPolyF2 res;
	int i;
	int j;
	int first_deg;
	int second_deg;
	int res_deg;
	unsigned char vals_first[TEST_SIZE_MAX];
	unsigned char vals_second[TEST_SIZE_MAX];
	unsigned char vals[TEST_SIZE_MAX];
	unsigned char * tmp = NULL;
	PolyType type;
	
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		type = PT_random();
		if(! make_alea_poly(&first, vals_first, type, first_deg = (1 + rand() % (TEST_SIZE(type) - 1))))
			return 0;
		type = PT_random();
		if(! make_alea_poly(&second, vals_second, type, second_deg = (1 + rand() % (TEST_SIZE(type) - 1))))
		{
			GenPolyF2_free(&first);
			return 0;
		}
		first_deg = GenPolyF2_real_degree(&first);
		second_deg = GenPolyF2_real_degree(&second);
		res_deg = (first_deg > second_deg) ? first_deg : second_deg;
		if(! (*function)(&res, &first, &second))
		{
			fprintf(stderr, "Add fail\n");
			GenPolyF2_free(&first);
			GenPolyF2_free(&second);
			return 0;
		}
		for(j = 0; j <= first_deg && j <= second_deg; ++j)
			vals[j] = vals_first[j] ^ vals_second[j];
		tmp = (first_deg > second_deg) ? vals_first : vals_second;
		for( ; j <= res_deg; ++j)
			vals[j] = *(tmp + j);
		tmp = NULL;
		if(! is_tab_equal(&res, vals, res_deg))
		{
			fprintf(stderr, "Incorrect values found\n");
			GenPolyF2_free(&first);
			GenPolyF2_free(&second);
			GenPolyF2_free(&res);
			return 0;
		}
		if(! is_tab_equal(&first, vals_first, first_deg)
		|| ! is_tab_equal(&second, vals_second, second_deg))
		{
			fprintf(stderr, "Modified values\n");
			GenPolyF2_free(&first);
			GenPolyF2_free(&second);
			GenPolyF2_free(&res);
			return 0;
		}
		GenPolyF2_free(&first);
		GenPolyF2_free(&second);
		GenPolyF2_free(&res);
	}
	
	return 1;
}

static int test_poly_op_add()
{
	return test_poly_op_add_aux(&GenPolyF2_add);
}

static int test_poly_op_sub()
{
	return test_poly_op_add_aux(&GenPolyF2_sub);
}

static int test_poly_op_mul_degree()
{
	GenPolyF2 poly;
	int deg;
	int add;
	unsigned char vals[TEST_SIZE_MAX];
	int i;
	int j;
	PolyType type;
	
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		type = PT_random();
		add = rand() % (TEST_SIZE(type) / 4);
		if(! make_alea_poly(&poly, vals, type, deg = (1 + rand() % (TEST_SIZE(type) - add - 1))))
			return 0;
		deg = GenPolyF2_real_degree(&poly);
		for(j = deg; j >= 0; --j)
			vals[j + add] = vals[j];
		for(j = add - 1; j >= 0; --j)
			vals[j] = 0;
		if(! GenPolyF2_mul_degree(&poly, add))
		{
			fprintf(stderr, "Mul degree fail\n");
			GenPolyF2_free(&poly);
			return 0;
		}
		if(! is_tab_equal(&poly, vals, deg))
		{
			fprintf(stderr, "Incorrect values\n");
			GenPolyF2_free(&poly);
			return 0;
		}
		GenPolyF2_free(&poly);
	}
	
	return 1;
}

static int test_poly_op_mul_simple()
{
	GenPolyF2 first;
	GenPolyF2 second;
	GenPolyF2 res;
	int i;
	int j;
	int k;
	int first_deg;
	int second_deg;
	int res_deg;
	unsigned char vals_first[TEST_SIZE_MAX];
	unsigned char vals_second[TEST_SIZE_MAX];
	unsigned char vals[TEST_SIZE_MAX];
	PolyType type;
	
	for(i = 0; i < TEST_TRIALS; ++i)
	{
		type = PT_random();
		if(! make_alea_poly(&first, vals_first, type, first_deg = (1 + rand() % (TEST_SIZE(type) / 2 - 1))))
			return 0;
		type = PT_random();
		if(! make_alea_poly(&second, vals_second, type, second_deg = (type == PT_INT) ? (1 + rand() % (TEST_SIZE(type) / 2 - 1)) : (1 + rand() % (TEST_SIZE(type) - first_deg - 1))))
		{
			GenPolyF2_free(&first);
			return 0;
		}
		first_deg = GenPolyF2_real_degree(&first);
		second_deg = GenPolyF2_real_degree(&second);
		res_deg = first_deg + second_deg;
		for(j = 0; j <= res_deg; ++j)
			vals[j] = 0;
		for(j = 0; j <= first_deg; ++j)
			if(vals_first[j])
				for(k = 0; k <= second_deg; ++k)
					vals[j + k] ^= vals_second[k];
		if(! GenPolyF2_mul(&res, &first, &second, NULL))
		{
			fprintf(stderr, "Mul fail\n");
			GenPolyF2_free(&first);
			GenPolyF2_free(&second);
			return 0;
		}
		if(! is_tab_equal(&res, vals, res_deg))
		{
			fprintf(stderr, "Incorrect values found\n");
			GenPolyF2_free(&first);
			GenPolyF2_free(&second);
			GenPolyF2_free(&res);
			return 0;
		}
		if(! is_tab_equal(&first, vals_first, first_deg)
		|| ! is_tab_equal(&second, vals_second, second_deg))
		{
			fprintf(stderr, "Modified values\n");
			GenPolyF2_free(&first);
			GenPolyF2_free(&second);
			GenPolyF2_free(&res);
			return 0;
		}
		GenPolyF2_free(&first);
		GenPolyF2_free(&second);
		GenPolyF2_free(&res);
	}
	
	return 1;
}

int test_poly_operations()
{
	Test list[] = {	{"Real degree", &test_real_degree},
			{"Poly add", &test_poly_op_add},
			{"Poly sub", &test_poly_op_sub},
			{"Poly mul degree", &test_poly_op_mul_degree},
			{"Poly mul simple", &test_poly_op_mul_simple}};
	const int list_len = 5;
	
	return make_test("Poly Operations", list, list_len, TM_NORMAL, STOP_AFTER_ERROR);
}

static void simul_error(GenPolyF2 * message, int nb_errors)
{
	int degree;
	int i;
	
	degree = GenPolyF2_real_degree(message);
	for(i = 0; i < nb_errors; ++i)
		GenPolyF2_add_val(message, 1, randint(0, degree));
}

int gen_test_BCH(GenPolyF2 * user_base)
{
	const int max_random_deg = 8;
	
	BCH code;
	GenPolyF2 based_poly;
	GenPolyF2 message;
	GenPolyF2 coded_message;
	GenPolyF2 final_message;
	GenPolyF2 * base = NULL;
	int random_deg;
	int nb_errors;
	int nb_correct;
	int nb_echec = 0;
	int nb_test = 0;
	
	srand(time(NULL));
	
	if(user_base != NULL)
	{
		if(! PrimitiveTest(user_base))
		{
			fprintf(stderr, "Erreur polynome non primitif\n");
			return 0;
		}
	}
	
	TC_init();
	while(TC_continue())
	{
		if(user_base == NULL)
		{
			random_deg = randint(4, max_random_deg);
			if(! GenPolyF2_load_random(&based_poly, random_deg))
			{
				fprintf(stderr, "Erreur %d irreductible non trouvé\n", random_deg);
				continue;
			}
			base = &based_poly;
			if(! PrimitiveTest(base))
				continue;
		}
		else
		{
			base = user_base;
			random_deg = GenPolyF2_real_degree(base);
		}
			
		nb_correct = randint(1, max_correct(random_deg));
		nb_errors = randint(0, nb_correct);
		if(! BCH_create(&code, base, nb_correct))
		{
			/*fprintf(stderr, "Erreur création BCH\n");*/
			if(base == &based_poly)
				GenPolyF2_free(base);
			continue;
		}
		if(code.msg_len == 1)
		{
			BCH_free(&code);
			if(base == &based_poly)
				GenPolyF2_free(base);
			continue;
		}
		BCH_print(&code);
		if(! make_alea_poly(&message, NULL, PT_NORMAL, code.msg_len - 1))
		{
			fprintf(stderr, "Erreur génération message taille %d\n", code.msg_len - 1);
			BCH_free(&code);
			if(base == &based_poly)
				GenPolyF2_free(base);
			continue;
		}
		if(! BCH_code(&code, &coded_message, &message))
		{
			fprintf(stderr, "Erreur codage BCH\n");
			GenPolyF2_free(&message);
			BCH_free(&code);
			if(base == &based_poly)
				GenPolyF2_free(base);
			continue;
		}
		simul_error(&coded_message, nb_errors);
		printf("%d erreurs générées\n", nb_errors);
		if(! BCH_decode(&code, &final_message, &coded_message))
		{
			fprintf(stderr, "Erreur décodage BCH\n");
			GenPolyF2_free(&coded_message);
			GenPolyF2_free(&message);
			BCH_free(&code);
			if(base == &based_poly)
				GenPolyF2_free(base);
			continue;
		}
		if(! GenPolyF2_equals(&final_message, &message))
		{
			++nb_echec;
			rgbprintf(TM_BOLD, RGB_BAD, "Erreur BCH(%d, %d, %d)\n", code.len, code.msg_len, code.nb_correct);
			GenPolyF2_print_bin(&message);
			GenPolyF2_print_bin(&final_message);
			GenPolyF2_self_add(&final_message, &message);
			printf("%d différences\n", Hamming_distance(&final_message));
			printf("reçu :\n");
			GenPolyF2_print_bin(&coded_message);
		}
		else
			rgbprintf(TM_BOLD, RGB_GOOD, "Success BCH(%d, %d, %d)\n", code.len, code.msg_len, code.nb_correct);
		GenPolyF2_free(&final_message);
		GenPolyF2_free(&coded_message);
		GenPolyF2_free(&message);
		BCH_free(&code);
		if(base == &based_poly)
			GenPolyF2_free(base);
		++nb_test;
	}
	TC_quit();
	if(nb_echec > 0)
		rgbprintf(TM_BOLD, RGB_BAD, "Erreur %d/%d echecs\n", nb_echec, nb_test);
	else
		rgbprintf(TM_BOLD, RGB_GOOD, "Success %d tests\n", nb_test);
	return nb_echec != 0;
}

int test_BCH()
{
	return gen_test_BCH(NULL);
}